from collections import OrderedDict
from enum import Enum
from functools import partial
from typing import List, Optional, Union
import numpy as np
import jax
import jax.numpy as jnp
from jax import lax
from jaxmarl.environments import MultiAgentEnv
from jaxmarl.environments import spaces
from typing import Tuple, Dict
import chex
from flax import struct
from flax.core.frozen_dict import FrozenDict
from jaxmarl.environments.overcooked_v2.common import (
    ACTION_TO_DIRECTION,
    MAX_INGREDIENTS,
    Actions,
    StaticObject,
    DynamicObject,
    Direction,
    Position,
    Agent,
)
from jaxmarl.environments.overcooked_v2.layouts import overcooked_v2_layouts, Layout
from jaxmarl.environments.overcooked_v2.settings import (
    DELIVERY_REWARD,
    INDICATOR_ACTIVATION_COST,
    INDICATOR_ACTIVATION_TIME,
    POT_COOK_TIME,
    SHAPED_REWARDS,
)
from jaxmarl.environments.overcooked_v2.utils import (
    OvercookedPathPlanner,
    compute_view_box,
    get_closest_true_pos_no_directions,
    mark_adjacent_cells,
    tree_select,
    compute_enclosed_spaces,
)


class ObservationType(str, Enum):
    DEFAULT = "default"
    FEATURIZED = "featurized"


@chex.dataclass
class State:
    agents: Agent

    # width x height x 3
    # First channel: static items
    # Second channel: dynamic items (plates and ingredients)
    # Third channel: extra info
    grid: chex.Array

    time: chex.Array
    terminal: bool

    recipe: int

    new_correct_delivery: bool = False

    ingredient_permutations: Optional[chex.Array] = None


class OvercookedV2(MultiAgentEnv):
    """Vanilla Overcooked"""

    def __init__(
        self,
        layout: Union[str, Layout] = "cramped_room",
        max_steps: int = 400,
        observation_type: Union[
            ObservationType, List[ObservationType]
        ] = ObservationType.DEFAULT,
        agent_view_size: Optional[int] = None,
        random_reset: bool = False,
        random_agent_positions: bool = False,
        start_cooking_interaction: bool = False,
        negative_rewards: bool = False,
        sample_recipe_on_delivery: bool = False,
        indicate_successful_delivery: bool = False,
        op_ingredient_permutations: List[int] = None,
        initial_state_buffer: Optional[State] = None,
        force_path_planning: bool = False,
    ):
        """
        Initializes the OvercookedV2 environment.

        Args:
            layout (Layout): The layout configuration for the environment, defaulting to "cramped_room". Either a Layout object or a string key to look up a Layout in overcooked_v2_layouts.
            max_steps (int): The maximum number of steps in the environment.
            observation_type (Union[ObservationType, List[ObservationType]]): The type of observation to use. Can be a single ObservationType or a list of ObservationTypes, one for each agent.
            agent_view_size (Optional[int]): The number of blocks the agent can view in each direction, None for full grid.
            random_reset (bool): Whether to reset the environment with random agent positions, inventories and pot states.
            random_agent_positions (bool): Whether to randomize agent positions. Agents will not be moved outside of their room if they are placed in an enclosed space.
            start_cooking_interaction (bool): If false the pot starts cooking automatically once three ingredients are added, if true the pot starts cooking only after the agent interacts with it.
            negative_rewards (bool): Whether to use negative rewards.
            sample_recipe_on_delivery (bool): Whether to sample a new recipe when a delivery is made. Default is on reset only.
            indicate_successful_delivery (bool): Whether to indicate a delivery was successful in the observation.
            op_ingredient_permutations (list): List of ingredient indices to permute in the observation (Fixed per agent in one episode).
            initial_state_buffer (State): Initial state buffer to be used to reset the environment. On each reset, a state from this buffer will be used.
            force_path_planning (bool): Whether to force path planning in the environment. Used to access featurized obs manually.
        """

        if isinstance(layout, str):
            if layout not in overcooked_v2_layouts:
                raise ValueError(
                    f"Invalid layout: {layout}, allowed layouts: {overcooked_v2_layouts.keys()}"
                )
            layout = overcooked_v2_layouts[layout]
        elif not isinstance(layout, Layout):
            raise ValueError("Invalid layout, must be a Layout object or a string key")

        num_agents = len(layout.agent_positions)

        super().__init__(num_agents=num_agents)

        self.height = layout.height
        self.width = layout.width

        self.layout = layout

        self.initial_state_buffer = initial_state_buffer

        self.agents = [f"agent_{i}" for i in range(num_agents)]
        self.action_set = jnp.array(list(Actions))

        if isinstance(observation_type, list):
            if len(observation_type) != num_agents:
                raise ValueError(
                    "Number of observation types must match the number of agents"
                )

        self.observation_type = observation_type

        self.agent_view_size = agent_view_size
        self.indicate_successful_delivery = indicate_successful_delivery
        self.obs_shape = self._get_obs_shape()

        self.max_steps = max_steps

        self.possible_recipes = jnp.array(layout.possible_recipes, dtype=jnp.int32)

        self.random_reset = random_reset
        self.random_agent_positions = random_agent_positions

        self.start_cooking_interaction = jnp.array(
            start_cooking_interaction, dtype=jnp.bool_
        )
        self.negative_rewards = negative_rewards
        self.sample_recipe_on_delivery = jnp.array(
            sample_recipe_on_delivery, dtype=jnp.bool_
        )

        self.enclosed_spaces = compute_enclosed_spaces(
            layout.static_objects == StaticObject.EMPTY,
        )

        self.op_ingredient_permutations = op_ingredient_permutations

        if (
            force_path_planning
            or observation_type == ObservationType.FEATURIZED
            or (
                isinstance(observation_type, list)
                and ObservationType.FEATURIZED in observation_type
            )
        ):
            move_area = layout.static_objects == StaticObject.EMPTY
            self.path_planer = OvercookedPathPlanner(move_area)

    def step_env(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: Dict[str, chex.Array],
    ) -> Tuple[Dict[str, chex.Array], State, Dict[str, float], Dict[str, bool], Dict]:
        """Perform single timestep state transition."""

        acts = self.action_set.take(
            indices=jnp.array([actions[f"agent_{i}"] for i in range(self.num_agents)])
        )

        state, reward, shaped_rewards = self.step_agents(key, state, acts)

        state = state.replace(time=state.time + 1)

        done = self.is_terminal(state)
        state = state.replace(terminal=done)

        obs = self.get_obs(state)

        rewards = {f"agent_{i}": reward for i in range(self.num_agents)}
        shaped_rewards = {
            f"agent_{i}": shaped_reward
            for i, shaped_reward in enumerate(shaped_rewards)
        }

        dones = {f"agent_{i}": done for i in range(self.num_agents)}
        dones["__all__"] = done

        return (
            lax.stop_gradient(obs),
            lax.stop_gradient(state),
            rewards,
            dones,
            {"shaped_reward": shaped_rewards},
        )

    @partial(jax.jit, static_argnums=(0,))
    def _sample_op_ingredient_permutations(self, key: chex.PRNGKey) -> chex.Array:
        perm_indices = jnp.array(self.op_ingredient_permutations)

        def _ingredient_permutation(key):
            full_perm = jnp.arange(self.layout.num_ingredients)
            perm = jax.random.permutation(key, perm_indices)
            full_perm = full_perm.at[perm_indices].set(full_perm[perm])
            return full_perm

        key, subkey = jax.random.split(key)
        ing_keys = jax.random.split(subkey, self.num_agents)
        ingredient_permutations = jax.vmap(_ingredient_permutation)(ing_keys)

        return ingredient_permutations

    def reset(
        self,
        key: chex.PRNGKey,
    ) -> Tuple[Dict[str, chex.Array], State]:
        if self.initial_state_buffer is not None:
            num_states = jax.tree_util.tree_flatten(self.initial_state_buffer)[0][
                0
            ].shape[0]
            # jax.debug.print("num_states: {i}", i=num_states)
            print("num_states in buffer: ", num_states)
            sampled_state_idx = jax.random.randint(key, (), 0, num_states)
            sampled_state = jax.tree_util.tree_map(
                lambda x: x[sampled_state_idx], self.initial_state_buffer
            )
            return self.reset_from_state(sampled_state, key)

        layout = self.layout

        static_objects = layout.static_objects
        grid = jnp.stack(
            [
                static_objects,
                jnp.zeros_like(static_objects),  # ingredient channel
                jnp.zeros_like(static_objects),  # extra info channel
            ],
            axis=-1,
            dtype=jnp.int32,
        )

        num_agents = self.num_agents
        x_positions, y_positions = map(jnp.array, zip(*layout.agent_positions))
        agents = Agent(
            pos=Position(x=x_positions, y=y_positions),
            dir=jnp.full((num_agents,), Direction.UP),
            inventory=jnp.zeros((num_agents,), dtype=jnp.int32),
        )

        key, subkey = jax.random.split(key)
        recipe = self._sample_recipe(subkey)

        ingredient_permutations = None
        if self.op_ingredient_permutations:
            ingredient_permutations = self._sample_op_ingredient_permutations(key)

        state = State(
            agents=agents,
            grid=grid,
            time=0,
            terminal=False,
            recipe=recipe,
            new_correct_delivery=False,
            ingredient_permutations=ingredient_permutations,
        )

        key, key_randomize = jax.random.split(key)
        if self.random_reset:
            state = self._randomize_state(state, key_randomize)
        elif self.random_agent_positions:
            state = self._randomize_agent_positions(state, key_randomize)

        obs = self.get_obs(state)

        return lax.stop_gradient(obs), lax.stop_gradient(state)

    @partial(jax.jit, static_argnums=(0,))
    def reset_from_state(
        self,
        state: State,
        key: chex.PRNGKey,
    ) -> Tuple[Dict[str, chex.Array], State]:
        """
        Reset the environment from a given state. Grid and agents are copied from the state, other parameters are reset.
        """

        print("reset_from_state")

        ingredient_permutations = None
        if self.op_ingredient_permutations:
            ingredient_permutations = self._sample_op_ingredient_permutations(key)

        state = state.replace(
            time=0,
            terminal=False,
            new_correct_delivery=False,
            ingredient_permutations=ingredient_permutations,
        )

        obs = self.get_obs(state)

        return lax.stop_gradient(obs), lax.stop_gradient(state)

    def _sample_recipe(self, key: chex.PRNGKey) -> int:
        fixed_recipe_idx = jax.random.randint(
            key, (), 0, self.possible_recipes.shape[0]
        )

        fixed_recipe = self.possible_recipes[fixed_recipe_idx]

        return DynamicObject.get_recipe_encoding(fixed_recipe)

    def _randomize_agent_positions(self, state: State, key: chex.PRNGKey) -> State:
        """Randomize agent positions."""
        num_agents = self.num_agents
        agents = state.agents

        def _select_agent_position(taken_mask, x):
            pos, key = x

            allowed_positions = (
                self.enclosed_spaces == self.enclosed_spaces[pos.y, pos.x]
            ) & ~taken_mask
            allowed_positions = allowed_positions.flatten()

            p = allowed_positions / jnp.sum(allowed_positions)
            agent_pos_idx = jax.random.choice(key, allowed_positions.size, (), p=p)
            agent_position = Position(
                x=agent_pos_idx % self.width, y=agent_pos_idx // self.width
            )

            new_taken_mask = taken_mask.at[agent_position.y, agent_position.x].set(True)
            return new_taken_mask, agent_position

        taken_mask = jnp.zeros_like(self.enclosed_spaces, dtype=jnp.bool_)
        key, subkey = jax.random.split(key)
        keys = jax.random.split(subkey, num_agents)
        _, agent_positions = jax.lax.scan(
            _select_agent_position, taken_mask, (agents.pos, keys)
        )

        # print("agent_positions: ", agent_positions)

        key, subkey = jax.random.split(key)
        directions = jax.random.randint(subkey, (num_agents,), 0, len(Direction))

        return state.replace(agents=agents.replace(pos=agent_positions, dir=directions))

    def _randomize_state(self, state: State, key: chex.PRNGKey) -> State:
        """Randomize the state of the environment."""

        key, subkey = jax.random.split(key)
        state = self._randomize_agent_positions(state, subkey)

        num_agents = self.num_agents
        agents = state.agents
        grid = state.grid

        # Agent inventory
        def _sample_inventory(key):
            key_dish, key_ing, key_inv = jax.random.split(key, 3)

            def _sample_dish(key):
                recipe_idx = jax.random.randint(key, (), 0, len(self.possible_recipes))
                recipe = self.possible_recipes[recipe_idx]
                return (
                    DynamicObject.get_recipe_encoding(recipe)
                    | DynamicObject.COOKED
                    | DynamicObject.PLATE
                )

            ingridient_idx = jax.random.randint(
                key_ing, (), 0, self.layout.num_ingredients
            )

            possible_inventory = jnp.array(
                [
                    DynamicObject.EMPTY,
                    DynamicObject.PLATE,
                    DynamicObject.ingredient(ingridient_idx),
                    _sample_dish(key_dish),
                ],
                dtype=jnp.int32,
            )

            inventory = jax.random.choice(
                key_inv, possible_inventory, (), p=jnp.array([0.5, 0.1, 0.25, 0.15])
            )
            return inventory

        key, subkey = jax.random.split(key)
        agent_inventories = jax.vmap(_sample_inventory)(
            jax.random.split(subkey, num_agents)
        )

        def _sample_grid_states_wrapper(cell, key):

            def _sample_pot_states(key):
                key, key_ing, key_num, key_timer = jax.random.split(key, 4)
                raw_ingridients = jax.random.randint(
                    key_ing, (3,), 0, self.layout.num_ingredients
                )
                raw_ingridients = jax.vmap(DynamicObject.ingredient)(raw_ingridients)

                partial_recipe = jax.random.randint(key_num, (), 1, 4)
                mask = jnp.arange(3) < partial_recipe

                pot_ingridients_masked = jnp.sum(raw_ingridients * mask)
                if self.start_cooking_interaction:
                    pot_ingridients_full = pot_ingridients_masked
                else:
                    # without an interaction the pot is always full when cooking
                    pot_ingridients_full = jnp.sum(raw_ingridients)

                pot_timer = jax.random.randint(key_timer, (), 0, POT_COOK_TIME) + 1

                possible_states = jnp.array(
                    [
                        cell,
                        [cell[0], pot_ingridients_masked, 0],
                        [cell[0], pot_ingridients_full, pot_timer],
                        [cell[0], pot_ingridients_full | DynamicObject.COOKED, 0],
                    ]
                )
                # 0 for do nothing, 1 for not started, 2 for started cooking, 3 for finished cooking
                return jax.random.choice(
                    key, possible_states, p=jnp.array([0.4, 0.35, 0.15, 0.1])
                )

            def _sample_counter_state(key):
                key, key_ing, key_dish = jax.random.split(key, 3)

                ingridient_idx = jax.random.randint(
                    key_ing, (), 0, self.layout.num_ingredients
                )
                dish_idx = jax.random.randint(
                    key_dish, (), 0, len(self.possible_recipes)
                )
                dish = (
                    DynamicObject.get_recipe_encoding(self.possible_recipes[dish_idx])
                    | DynamicObject.COOKED
                    | DynamicObject.PLATE
                )

                possible_states = jnp.array(
                    [
                        DynamicObject.EMPTY,
                        DynamicObject.PLATE,
                        DynamicObject.ingredient(ingridient_idx),
                        dish,
                    ]
                )

                ing_layer = jax.random.choice(
                    key, possible_states, p=jnp.array([0.5, 0.1, 0.3, 0.1])
                )
                return cell.at[1].set(ing_layer)

            is_pot = cell[0] == StaticObject.POT
            is_wall = cell[0] == StaticObject.WALL
            branch_idx = 1 * is_pot + 2 * is_wall

            return jax.lax.switch(
                branch_idx,
                [
                    lambda _: cell,
                    lambda key: _sample_pot_states(key),
                    lambda key: _sample_counter_state(key),
                ],
                key,
            )

        key, subkey = jax.random.split(key)
        key_grid = jax.random.split(subkey, (self.height, self.width))
        new_grid = jax.vmap(jax.vmap(_sample_grid_states_wrapper))(grid, key_grid)

        # print("new_grid: ", new_grid)

        return state.replace(
            agents=agents.replace(inventory=agent_inventories),
            grid=new_grid,
        )

    def _get_obs_shape(self) -> Tuple[int]:
        if self.agent_view_size:
            view_size = self.agent_view_size * 2 + 1
            view_width = min(self.width, view_size)
            view_height = min(self.height, view_size)
        else:
            view_width = self.width
            view_height = self.height

        def _get_obs_shape_single(obs_type):
            match obs_type:
                case ObservationType.DEFAULT:
                    num_ingredients = self.layout.num_ingredients
                    num_layers = 18 + 4 * (num_ingredients + 2)

                    if self.indicate_successful_delivery:
                        num_layers += 1

                    return (view_height, view_width, num_layers)
                case ObservationType.FEATURIZED:
                    num_pot_features = 10
                    base_features = 28

                    # TODO: maybe pass this as argument
                    num_pots = 2
                    total_features = self.num_agents * (
                        num_pots * num_pot_features + base_features
                    )
                    return (total_features,)
                case _:
                    raise ValueError(
                        f"Invalid observation type: {self.observation_type}"
                    )

        if isinstance(self.observation_type, list):
            return [
                _get_obs_shape_single(obs_type) for obs_type in self.observation_type
            ]

        return _get_obs_shape_single(self.observation_type)

    def get_obs(self, state: State) -> chex.Array:
        if not isinstance(self.observation_type, list):
            return self.get_obs_for_type(state, self.observation_type)

        all_obs = {}
        for i, obs_type in enumerate(self.observation_type):
            obs = self.get_obs_for_type(state, obs_type)
            key = f"agent_{i}"
            all_obs[key] = obs[key]
        return all_obs

    def get_obs_for_type(
        self, state: State, obs_type: ObservationType
    ) -> Dict[str, chex.Array]:
        match obs_type:
            case ObservationType.DEFAULT:
                all_obs = self.get_obs_default(state)
            case ObservationType.FEATURIZED:
                all_obs = self.get_obs_featurized(state)
            case _:
                raise ValueError(f"Invalid observation type: {self.observation_type}")

        def _mask_obs(obs, agent):
            view_size = self.agent_view_size
            pos = agent.pos

            padded_obs = jnp.pad(
                obs,
                ((view_size, view_size), (view_size, view_size), (0, 0)),
                mode="constant",
                constant_values=0,
            )

            sliced_obs = jax.lax.dynamic_slice(
                padded_obs,
                (pos.y, pos.x, 0),
                self.obs_shape,
            )

            return sliced_obs

        if self.agent_view_size is not None:
            all_obs = jax.vmap(_mask_obs)(all_obs, state.agents)

        return {f"agent_{i}": obs for i, obs in enumerate(all_obs)}

    def get_obs_default(self, state: State) -> Dict[str, chex.Array]:

        width = self.width
        height = self.height
        num_ingredients = self.layout.num_ingredients

        static_objects = state.grid[:, :, 0]
        ingredients = state.grid[:, :, 1]
        extra_info = state.grid[:, :, 2]

        static_encoding = jnp.array(
            [
                StaticObject.WALL,
                StaticObject.GOAL,
                StaticObject.POT,
                StaticObject.RECIPE_INDICATOR,
                StaticObject.BUTTON_RECIPE_INDICATOR,
                StaticObject.PLATE_PILE,
            ]
        )
        static_layers = static_objects[..., None] == static_encoding
        # print("static_layers: ", static_layers.shape)

        def _ingridient_layers(ingredients, ingredient_mapping=None):
            shift = jnp.array([0, 1] + [2 * (i + 1) for i in range(num_ingredients)])
            mask = jnp.array([0x1, 0x1] + [0x3] * num_ingredients)

            layers = ingredients[..., None] >> shift
            layers = layers & mask

            if ingredient_mapping is not None:
                layers = layers.at[..., 2:].set(
                    layers[..., 2:][..., ingredient_mapping]
                )

            return layers

        # ingredients_layers = _ingridient_layers(ingredients)
        # print("ingredients_layers: ", ingredients_layers.shape)

        recipe_indicator_mask = static_objects == StaticObject.RECIPE_INDICATOR
        button_recipe_indicator_mask = (
            static_objects == StaticObject.BUTTON_RECIPE_INDICATOR
        ) & (extra_info > 0)

        recipe_ingridients = jnp.where(
            recipe_indicator_mask | button_recipe_indicator_mask, state.recipe, 0
        )
        # recipe_layers = _ingridient_layers(recipe_ingridients)
        # print("recipe_layers: ", recipe_layers.shape)

        extra_info = state.grid[:, :, 2]
        pot_timer_layer = jnp.where(static_objects == StaticObject.POT, extra_info, 0)
        new_correct_delivery_layer = jnp.where(
            static_objects == StaticObject.GOAL, state.new_correct_delivery, 0
        )

        extra_layers = [
            pot_timer_layer,
        ]

        if self.indicate_successful_delivery:
            extra_layers.append(new_correct_delivery_layer)

        extra_layers = jnp.stack(extra_layers, axis=-1)

        def _agent_layers(agent, ingredient_mapping=None):
            pos = agent.pos
            direction = agent.dir
            inv = agent.inventory

            pos_layers = (
                jnp.zeros((height, width, 1), dtype=jnp.uint8)
                .at[pos.y, pos.x, 0]
                .set(1)
            )
            dir_layers = (
                jnp.zeros((height, width, 4), dtype=jnp.uint8)
                .at[pos.y, pos.x, direction]
                .set(1)
            )
            inv_grid = jnp.zeros_like(ingredients).at[pos.y, pos.x].set(inv)
            inv_layers = _ingridient_layers(
                inv_grid, ingredient_mapping=ingredient_mapping
            )

            return jnp.concatenate(
                [
                    pos_layers,
                    dir_layers,
                    inv_layers,
                ],
                axis=-1,
            )



        def _agent_obs(agent_id):
            ingredient_mapping = None
            if self.op_ingredient_permutations:
                ingredient_mapping = state.ingredient_permutations[agent_id]

            agent_layers = jax.vmap(
                partial(_agent_layers, ingredient_mapping=ingredient_mapping)
            )(state.agents)
            agent_layer = agent_layers[agent_id]
            all_agent_layers = jnp.sum(agent_layers, axis=0)

            other_agent_layers = all_agent_layers - agent_layer
            # print("agent_layer: ", agent_layer.shape)

            ingredients_layers = _ingridient_layers(
                ingredients, ingredient_mapping=ingredient_mapping
            )
            # print("ingredients_layers: ", ingredients_layers.shape)

            recipe_layers = _ingridient_layers(
                recipe_ingridients, ingredient_mapping=ingredient_mapping
            )
            # print("recipe_layers: ", recipe_layers.shape)

            ingredient_pile_encoding = jnp.array(
                [StaticObject.INGREDIENT_PILE_BASE + i for i in range(num_ingredients)]
            )
            if self.op_ingredient_permutations:
                ingredient_pile_encoding = ingredient_pile_encoding[ingredient_mapping]

            ingredient_pile_layers = (
                static_objects[..., None] == ingredient_pile_encoding
            )

            return jnp.concatenate(
                [
                    agent_layer,
                    other_agent_layers,
                    static_layers,
                    ingredient_pile_layers,
                    ingredients_layers,
                    recipe_layers,
                    extra_layers,
                ],
                axis=-1,
            )

        return jax.vmap(_agent_obs)(jnp.arange(self.num_agents))

    def get_obs_featurized(self, state: State) -> chex.Array:
        """
        Observation to match featurized observation from OvercookedAI, this method is used to featurize the state of the environment.
        Encode state with some manually designed features. Works for arbitrary number of players

        Arguments:
            overcooked_state (OvercookedState): state we wish to featurize
            num_pots (int): Encode the state (ingredients, whether cooking or not, etc) of the 'num_pots' closest pots to each player.
                If i < num_pots pots are reachable by player i, then pots [i+1, num_pots] are encoded as all zeros. Changing this
                impacts the shape of the feature encoding

        Returns:
            ordered_features (list[np.Array]): The ith element contains a player-centric featurized view for the ith player

            The encoding for player i is as follows:

                [player_i_features, other_player_features player_i_dist_to_other_players, player_i_position]

                player_{i}_features (length num_pots*10 + 28):
                    pi_orientation: length 4 one-hot-encoding of direction currently facing
                    pi_obj: length 4 one-hot-encoding of object currently being held (all 0s if no object held)
                    pi_closest_{onion|tomato|dish|soup|serving|empty_counter}: (dx, dy) where dx = x dist to item, dy = y dist to item. (0, 0) if item is currently held
                    pi_cloest_soup_n_{onions|tomatoes}: int value for number of this ingredient in closest soup
                    pi_closest_pot_{j}_exists: {0, 1} depending on whether jth closest pot found. If 0, then all other pot features are 0. Note: can
                        be 0 even if there are more than j pots on layout, if the pot is not reachable by player i
                    pi_closest_pot_{j}_{is_empty|is_full|is_cooking|is_ready}: {0, 1} depending on boolean value for jth closest pot
                    pi_closest_pot_{j}_{num_onions|num_tomatoes}: int value for number of this ingredient in jth closest pot
                    pi_closest_pot_{j}_cook_time: int value for seconds remaining on soup. -1 if no soup is cooking
                    pi_closest_pot_{j}: (dx, dy) to jth closest pot from player i location
                    pi_wall: length 4 boolean value of whether player i has wall in each direction

                other_player_features (length (num_players - 1)*(num_pots*10 + 24)):
                    ordered concatenation of player_{j}_features for j != i

                player_i_dist_to_other_players (length (num_players - 1)*2):
                    [player_j.pos - player_i.pos for j != i]

                player_i_position (length 2)
        """
        if self.layout.num_ingredients > 1:
            # environment dynamics with two ingredients differs to the original OvercookedAI
            raise NotImplementedError(
                "Featurized observation not implemented for more than 1 ingredient"
            )

        # TODO: maybe pass as argument
        num_pots = 2
        ignore_counters = False

        onion = DynamicObject.ingredient(0)
        recipe = 3 * onion
        soup = recipe | DynamicObject.COOKED | DynamicObject.PLATE

        move_area = state.grid[:, :, 0] == StaticObject.EMPTY

        def _player_features(agent):
            pos = agent.pos
            direction = agent.dir
            inv = agent.inventory

            reachable_area = self.enclosed_spaces == self.enclosed_spaces[pos.y, pos.x]
            reachable_area = mark_adjacent_cells(reachable_area)

            # pi_orientation: [NORTH, SOUTH, EAST, WEST]
            dir_features = jax.nn.one_hot(direction, 4)

            # pi_obj: ["onion", "soup", "dish", "tomato"]
            items = jnp.array(
                [
                    DynamicObject.ingredient(0),
                    soup,
                    DynamicObject.PLATE,
                    DynamicObject.ingredient(1),
                ]
            )
            inv_features = inv[..., None] == items

            def _closest_features(
                static_locator=None,
                dynamic_locator=None,
                no_ingredients=False,
                not_in_pot=True,
            ):
                mask = jnp.zeros((self.height, self.width), dtype=jnp.bool_)
                if static_locator:
                    static_mask = state.grid[:, :, 0] == static_locator
                    if no_ingredients:
                        static_mask &= state.grid[:, :, 1] == DynamicObject.EMPTY
                    mask |= static_mask
                if dynamic_locator:
                    dynamic_mask = state.grid[:, :, 1] == dynamic_locator
                    if not_in_pot:
                        dynamic_mask &= state.grid[:, :, 0] != StaticObject.POT
                    if ignore_counters:
                        dynamic_mask &= state.grid[:, :, 0] != StaticObject.WALL
                    mask |= dynamic_mask
                    mask = mask.at[pos.y, pos.x].set(inv == dynamic_locator)

                mask &= reachable_area


                obj_pos, is_valid = self.path_planer.get_closest_target_pos(
                    mask, pos, direction
                )

                delta = obj_pos.delta(pos)
                return jax.lax.select(is_valid, delta, jnp.array([0, 0]))

            # pi_closest_{onion|tomato|dish|soup|serving|empty_counter}
            onion_features = _closest_features(
                static_locator=StaticObject.ingredient_pile(0),
                dynamic_locator=DynamicObject.ingredient(0),
            )
            tomato_features = _closest_features(
                static_locator=StaticObject.ingredient_pile(1),
                dynamic_locator=DynamicObject.ingredient(1),
            )
            dish_features = _closest_features(
                static_locator=StaticObject.PLATE_PILE,
                dynamic_locator=DynamicObject.PLATE,
            )
            soup_features = _closest_features(dynamic_locator=soup)
            serving_features = _closest_features(static_locator=StaticObject.GOAL)
            empty_counter_features = _closest_features(
                static_locator=StaticObject.WALL, no_ingredients=True
            )
            if ignore_counters:
                empty_counter_features = jnp.array([0, 0])

            # pi_closest_soup_n_{onions|tomatoes}
            # we assume that recipe is always 3 onions
            soup_on_grid_mask = state.grid[:, :, 1] == soup
            if ignore_counters:
                soup_on_grid_mask &= state.grid[:, :, 0] != StaticObject.WALL
            soup_onions = jax.lax.select(
                jnp.any(soup_on_grid_mask) | (inv == soup), 3, 0
            )
            soup_tomatoes = 0
            soup_ingredient_features = jnp.array([soup_onions, soup_tomatoes])

            def _compute_pot_features(agent, pot_pos, pot_ing, pot_timer):
                # pi_closest_pot_{j}_exists
                pot_exists = 1

                # pi_closest_pot_{j}_{is_empty|is_full|is_cooking|is_ready}
                pot_empty = pot_ing == DynamicObject.EMPTY
                pot_full = DynamicObject.ingredient_count(pot_ing) == 3
                pot_cooking = pot_timer > 0
                pot_ready = (pot_ing & DynamicObject.COOKED) == DynamicObject.COOKED

                # pi_closest_pot_{j}_{num_onions|num_tomatoes}
                num_onions = DynamicObject.ingredient_count(pot_ing)
                num_tomatoes = 0

                # pi_closest_pot_{j}_cook_time
                cook_time = pot_timer

                # pi_closest_pot_{j}
                pot_deltas = pot_pos.delta(agent.pos)

                pot_features = jnp.array(
                    [
                        pot_exists,
                        pot_empty,
                        pot_full,
                        pot_cooking,
                        pot_ready,
                        num_onions,
                        num_tomatoes,
                        cook_time,
                        *pot_deltas,
                    ]
                )
                return pot_features

            pot_mask = state.grid[:, :, 0] == StaticObject.POT
            pot_mask &= reachable_area

            all_pot_features = jnp.zeros((num_pots, 10))
            for i in range(num_pots):

                def _process_pot(pos):
                    y, x = pos.y, pos.x
                    pot_ing = state.grid[y, x, 1]
                    pot_timer = state.grid[y, x, 2]

                    return _compute_pot_features(agent, pos, pot_ing, pot_timer)

                pot_pos, is_valid = self.path_planer.get_closest_target_pos(
                    pot_mask, pos, direction
                )

                pot_features = jax.lax.cond(
                    is_valid,
                    _process_pot,
                    lambda _: jnp.zeros(10, dtype=jnp.int32),
                    pot_pos,
                )
                pot_mask = pot_mask.at[pot_pos.y, pot_pos.x].set(False)

                all_pot_features = all_pot_features.at[i].set(pot_features)

            all_pot_features = all_pot_features.flatten()

            # pi_wall: [NORTH, SOUTH, EAST, WEST]
            wall_mask = ~move_area
            wall_features = jnp.array(
                [
                    wall_mask[pos.y - 1, pos.x],
                    wall_mask[pos.y + 1, pos.x],
                    wall_mask[pos.y, pos.x + 1],
                    wall_mask[pos.y, pos.x - 1],
                ]
            )

            return jnp.concatenate(
                [
                    dir_features,
                    inv_features,
                    onion_features,
                    tomato_features,
                    dish_features,
                    soup_features,
                    soup_ingredient_features,
                    serving_features,
                    empty_counter_features,
                    all_pot_features,
                    wall_features,
                ]
            )

        all_player_features = jax.vmap(_player_features)(state.agents)

        def _agent_obs(agent, i):
            agent_features = all_player_features[i]

            other_agent_selector = jnp.arange(self.num_agents - 1)
            other_agent_selector += other_agent_selector >= i
            # print("other_agent_selector: ", other_agent_selector)

            other_player_features = jnp.concatenate(
                all_player_features[other_agent_selector], axis=-1
            )

            # print("agent_features: ", agent_features)
            # print("other_player_features: ", other_player_features)

            def _dist_to_other_players(other_agent):
                return other_agent.pos.delta(agent.pos)

            dist_to_other_players = jax.vmap(_dist_to_other_players)(state.agents)
            dist_to_other_players = dist_to_other_players[
                other_agent_selector
            ].flatten()

            # print("dist_to_other_players: ", dist_to_other_players)
            # print("agent_pos: ", agent.pos.to_array())

            return jnp.concatenate(
                [
                    agent_features,
                    other_player_features,
                    dist_to_other_players,
                    agent.pos.to_array(),
                ],
                axis=-1,
            )

        return jax.vmap(_agent_obs)(state.agents, jnp.arange(self.num_agents))

    def step_agents(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: chex.Array,
    ) -> Tuple[State, float]:
        grid = state.grid

        # print("actions: ", actions)

        # Move action:
        # 1. move agent to new position (if possible on the grid)
        # 2. resolve collisions
        # 3. prevent swapping
        def _move_wrapper(agent, action):
            direction = ACTION_TO_DIRECTION[action]

            def _move(agent, dir):
                pos = agent.pos
                new_pos = pos.move_in_bounds(dir, self.width, self.height)

                new_pos = tree_select(
                    grid[new_pos.y, new_pos.x, 0] == StaticObject.EMPTY, new_pos, pos
                )

                return agent.replace(pos=new_pos, dir=direction)

            return jax.lax.cond(
                direction != -1,
                _move,
                lambda a, _: a,
                agent,
                direction,
            )

        new_agents = jax.vmap(_move_wrapper)(state.agents, actions)

        # Resolve collisions:
        def _masked_positions(mask):
            return tree_select(mask, state.agents.pos, new_agents.pos)

        def _get_collisions(mask):
            positions = _masked_positions(mask)

            collision_grid = jnp.zeros((self.height, self.width))
            collision_grid, _ = jax.lax.scan(
                lambda grid, pos: (grid.at[pos.y, pos.x].add(1), None),
                collision_grid,
                positions,
            )

            collision_mask = collision_grid > 1

            collisions = jax.vmap(lambda p: collision_mask[p.y, p.x])(positions)
            return collisions

        initial_mask = jnp.zeros((self.num_agents,), dtype=bool)
        mask = jax.lax.while_loop(
            lambda mask: jnp.any(_get_collisions(mask)),
            lambda mask: mask | _get_collisions(mask),
            initial_mask,
        )
        new_agents = new_agents.replace(pos=_masked_positions(mask))

        # Prevent swapping:
        def _compute_swapped_agents(original_positions, new_positions):
            original_positions = original_positions.to_array()
            new_positions = new_positions.to_array()

            original_pos_expanded = jnp.expand_dims(original_positions, axis=0)
            new_pos_expanded = jnp.expand_dims(new_positions, axis=1)

            swap_mask = (original_pos_expanded == new_pos_expanded).all(axis=-1)
            swap_mask = jnp.fill_diagonal(swap_mask, False, inplace=False)

            swap_pairs = jnp.logical_and(swap_mask, swap_mask.T)

            swapped_agents = jnp.any(swap_pairs, axis=0)
            return swapped_agents

        swap_mask = _compute_swapped_agents(state.agents.pos, new_agents.pos)
        new_agents = new_agents.replace(pos=_masked_positions(swap_mask))

        # Interact action:
        def _interact_wrapper(carry, x):
            agent, action = x
            is_interact = action == Actions.interact

            def _interact(carry, agent):
                grid, correct_delivery, reward = carry

                (
                    new_grid,
                    new_agent,
                    new_correct_delivery,
                    interact_reward,
                    shaped_reward,
                ) = self.process_interact(
                    grid, agent, new_agents.inventory, state.recipe
                )

                carry = (
                    new_grid,
                    correct_delivery | new_correct_delivery,
                    reward + interact_reward,
                )
                return carry, (new_agent, shaped_reward)

            return jax.lax.cond(
                is_interact, _interact, lambda c, a: (c, (a, 0.0)), carry, agent
            )

        carry = (grid, False, 0.0)
        xs = (new_agents, actions)
        (new_grid, new_correct_delivery, reward), (new_agents, shaped_rewards) = (
            jax.lax.scan(_interact_wrapper, carry, xs)
        )

        # Update extra info:
        def _timestep_wrapper(cell):
            def _cook(cell):
                is_cooking = cell[2] > 0
                new_extra = jax.lax.select(is_cooking, cell[2] - 1, cell[2])
                finished_cooking = is_cooking * (new_extra == 0)
                new_ingredients = cell[1] | (finished_cooking * DynamicObject.COOKED)

                return jnp.array([cell[0], new_ingredients, new_extra])

            def _indicator(cell):
                new_extra = jnp.clip(cell[2] - 1, min=0)
                return cell.at[2].set(new_extra)

            # return jax.lax.cond(is_pot, _cook, lambda x: x, cell)

            branches = (
                jnp.array(
                    [
                        StaticObject.POT,
                        StaticObject.BUTTON_RECIPE_INDICATOR,
                    ]
                )
                == cell[0]
            )

            branch_idx = jax.lax.select(
                jnp.any(branches),
                jnp.argmax(branches) + 1,
                0,
            )

            return jax.lax.switch(
                branch_idx,
                [
                    lambda x: x,
                    _cook,
                    _indicator,
                ],
                cell,
            )

        new_grid = jax.vmap(jax.vmap(_timestep_wrapper))(new_grid)

        sample_new_recipe = new_correct_delivery & self.sample_recipe_on_delivery

        key, subkey = jax.random.split(key)
        new_recipe = jax.lax.cond(
            sample_new_recipe,
            lambda _, key: self._sample_recipe(key),
            lambda r, _: r,
            state.recipe,
            subkey,
        )

        return (
            state.replace(
                agents=new_agents,
                grid=new_grid,
                recipe=new_recipe,
                new_correct_delivery=new_correct_delivery,
            ),
            reward,
            shaped_rewards,
        )

    def process_interact(
        self,
        grid: chex.Array,
        agent: Agent,
        all_inventories: jnp.ndarray,
        recipe: int,
    ):
        """Assume agent took interact actions. Result depends on what agent is facing and what it is holding."""

        inventory = agent.inventory
        fwd_pos = agent.get_fwd_pos()

        shaped_reward = jnp.array(0, dtype=float)

        interact_cell = grid[fwd_pos.y, fwd_pos.x]

        interact_item = interact_cell[0]
        interact_ingredients = interact_cell[1]
        interact_extra = interact_cell[2]
        plated_recipe = recipe | DynamicObject.PLATE | DynamicObject.COOKED

        # Booleans depending on what the object is
        object_is_plate_pile = interact_item == StaticObject.PLATE_PILE
        object_is_ingredient_pile = StaticObject.is_ingredient_pile(interact_item)

        object_is_pile = object_is_plate_pile | object_is_ingredient_pile
        object_is_pot = interact_item == StaticObject.POT
        object_is_goal = interact_item == StaticObject.GOAL
        object_is_wall = interact_item == StaticObject.WALL
        onject_is_button_recipe_indicator = (
            interact_item == StaticObject.BUTTON_RECIPE_INDICATOR
        )

        object_has_no_ingredients = interact_ingredients == 0

        inventory_is_empty = inventory == 0
        inventory_is_ingredient = DynamicObject.is_ingredient(inventory)
        # print("inventory_is_ingredient: ", inventory_is_ingredient)
        inventory_is_plate = inventory == DynamicObject.PLATE
        inventory_is_dish = (inventory & DynamicObject.COOKED) != 0

        merged_ingredients = interact_ingredients + inventory
        # print("merged_ingredients: ", merged_ingredients)

        pot_is_cooking = object_is_pot * (interact_extra > 0)
        pot_is_cooked = object_is_pot * (
            interact_ingredients & DynamicObject.COOKED != 0
        )
        pot_is_idle = object_is_pot * ~pot_is_cooking * ~pot_is_cooked

        successful_dish_pickup = pot_is_cooked * inventory_is_plate
        is_dish_pickup_useful = merged_ingredients == plated_recipe
        shaped_reward += (
            successful_dish_pickup
            * is_dish_pickup_useful
            * SHAPED_REWARDS["DISH_PICKUP"]
        )

        successful_pickup = (
            object_is_pile * inventory_is_empty
            + successful_dish_pickup
            + object_is_wall * ~object_has_no_ingredients * inventory_is_empty
        )

        successful_indicator_activation = (
            onject_is_button_recipe_indicator
            * inventory_is_empty
            * object_has_no_ingredients
        )

        # print("successful_pickup: ", successful_pickup)
        # print("object_is_pile: ", object_is_pile)
        # print("inventory_is_empty: ", inventory_is_empty)

        pot_full = DynamicObject.ingredient_count(interact_ingredients) == 3
        # print("pot_full: ", pot_full)

        successful_pot_placement = pot_is_idle * inventory_is_ingredient * ~pot_full
        ingredient_selector = inventory | (inventory << 1)
        is_pot_placement_useful = (interact_ingredients & ingredient_selector) < (
            recipe & ingredient_selector
        )
        shaped_reward += (
            successful_pot_placement
            * is_pot_placement_useful
            * jax.lax.select(
                is_pot_placement_useful,
                1,
                -1 if self.negative_rewards else 0,
            )
            * SHAPED_REWARDS["PLACEMENT_IN_POT"]
        )

        successful_drop = (
            object_is_wall * object_has_no_ingredients * ~inventory_is_empty
            + successful_pot_placement
        )
        successful_delivery = object_is_goal * inventory_is_dish
        no_effect = ~successful_pickup * ~successful_drop * ~successful_delivery

        pile_ingredient = (
            object_is_plate_pile * DynamicObject.PLATE
            + object_is_ingredient_pile * StaticObject.get_ingredient(interact_item)
        )
        # print("pile_ingredient: ", pile_ingredient)

        new_ingredients = (
            successful_drop * merged_ingredients + no_effect * interact_ingredients
        )
        pot_full_after_drop = DynamicObject.ingredient_count(new_ingredients) == 3

        successful_pot_start_cooking = (
            pot_is_idle
            * ~object_has_no_ingredients
            * inventory_is_empty
            * self.start_cooking_interaction
        )
        is_pot_start_cooking_useful = interact_ingredients == recipe
        shaped_reward += (
            successful_pot_start_cooking
            * is_pot_start_cooking_useful
            * SHAPED_REWARDS["POT_START_COOKING"]
        )
        auto_cook = pot_is_idle & pot_full_after_drop & ~self.start_cooking_interaction

        use_pot_extra = successful_pot_start_cooking | auto_cook
        new_extra = (
            use_pot_extra * POT_COOK_TIME
            + successful_indicator_activation * INDICATOR_ACTIVATION_TIME
            + ~use_pot_extra * ~successful_indicator_activation * interact_extra
        )

        new_cell = jnp.array([interact_item, new_ingredients, new_extra])

        new_grid = grid.at[fwd_pos.y, fwd_pos.x].set(new_cell)

        new_inventory = (
            successful_pickup * (pile_ingredient + merged_ingredients)
            + no_effect * inventory
        )
        # print("new_inventory: ", new_inventory)
        new_agent = agent.replace(inventory=new_inventory)

        is_correct_recipe = inventory == plated_recipe
        # print("is_correct_recipe: ", is_correct_recipe)

        reward = jnp.array(0, dtype=float)

        # Reward for successful delivery of a dish (negative if the dish is incorrect)
        reward += (
            successful_delivery
            * jax.lax.select(
                is_correct_recipe,
                1,
                -1 if self.negative_rewards else 0,
            )
            * DELIVERY_REWARD
        )

        # Cost for activating a button recipe indicator
        reward -= successful_indicator_activation * INDICATOR_ACTIVATION_COST

        # Plate pickup reward: number of plates in player hands < number ready/cooking/partially full pot
        inventory_is_plate = new_inventory == DynamicObject.PLATE
        successful_plate_pickup = successful_pickup * inventory_is_plate
        num_plates_in_inventory = jnp.sum(all_inventories == DynamicObject.PLATE)
        num_nonempty_pots = jnp.sum(
            (grid[:, :, 0] == StaticObject.POT) & (grid[:, :, 1] != 0)
        )
        is_plate_pickup_useful = num_plates_in_inventory < num_nonempty_pots
        # make sure there are no plates on counters to prevent reward hacking
        no_plates_on_counters = jnp.sum(grid[:, :, 1] == DynamicObject.PLATE) == 0
        shaped_reward += (
            no_plates_on_counters
            * is_plate_pickup_useful
            * successful_plate_pickup
            * SHAPED_REWARDS["PLATE_PICKUP"]
        )

        correct_delivery = successful_delivery & is_correct_recipe
        return new_grid, new_agent, correct_delivery, reward, shaped_reward

    def is_terminal(self, state: State) -> bool:
        """Check whether state is terminal."""
        done_steps = state.time >= self.max_steps
        return done_steps | state.terminal

    def get_eval_solved_rate_fn(self):
        def _fn(ep_stats):
            return ep_stats["return"] > 0

        return _fn

    @property
    def name(self) -> str:
        """Environment name."""
        return "Overcooked V2"

    @property
    def num_actions(self) -> int:
        """Number of actions possible in environment."""
        return len(self.action_set)

    def action_space(self, agent_id="") -> spaces.Discrete:
        """Action space of the environment. Agent_id not used since action_space is uniform for all agents"""
        return spaces.Discrete(len(self.action_set), dtype=jnp.uint32)

    def observation_space(self) -> spaces.Box:
        """Observation space of the environment."""
        return spaces.Box(0, 255, self.obs_shape)

    def state_space(self) -> spaces.Dict:
        """State space of the environment."""
        h = self.height
        w = self.width
        agent_view_size = self.agent_view_size
        return spaces.Dict(
            {
                "agent_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
                "agent_dir": spaces.Discrete(4),
                "goal_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
                "maze_map": spaces.Box(
                    0,
                    255,
                    (w + agent_view_size, h + agent_view_size, 3),
                    dtype=jnp.uint32,
                ),
                "time": spaces.Discrete(self.max_steps),
                "terminal": spaces.Discrete(2),
            }
        )



"""
Short introduction to running the Overcooked environment and visualising it using random actions.
"""

import jax 
from jaxmarl import make
from jaxmarl.viz.overcooked_visualizer import OvercookedVisualizer
from jaxmarl.environments.overcooked import Overcooked, overcooked_layouts, layout_grid_to_dict
import time

# Parameters + random keys
max_steps = 100
key = jax.random.PRNGKey(0)
key, key_r, key_a = jax.random.split(key, 3)

# Get one of the classic layouts (cramped_room, asymm_advantages, coord_ring, forced_coord, counter_circuit)
layout = overcooked_layouts["cramped_room"]

# Or make your own!
# custom_layout_grid = """
# WWOWW
# WA  W
# B P X
# W  AW
# WWOWW
# """
# layout = layout_grid_to_dict(custom_layout_grid)

# Instantiate environment
env = make('overcooked', layout=layout, max_steps=max_steps)

obs, state = env.reset(key_r)
print('list of agents in environment', env.agents)

# Sample random actions
key_a = jax.random.split(key_a, env.num_agents)
actions = {agent: env.action_space(agent).sample(key_a[i]) for i, agent in enumerate(env.agents)}
print('example action dict', actions)

state_seq = []
for _ in range(max_steps):
    state_seq.append(state)
    # Iterate random keys and sample actions
    key, key_s, key_a = jax.random.split(key, 3)
    key_a = jax.random.split(key_a, env.num_agents)

    actions = {agent: env.action_space(agent).sample(key_a[i]) for i, agent in enumerate(env.agents)}

    # Step environment
    obs, state, rewards, dones, infos = env.step(key_s, state, actions)

viz = OvercookedVisualizer()

# Render to screen
for s in state_seq:
    viz.render(env.agent_view_size, s, highlight=False)
    time.sleep(0.25)

# # Or save an animation
# viz.animate(state_seq, agent_view_size=5, filename='animation.gif')


import copy

import numpy as np
import jax.numpy as jnp
from flax import struct
import chex
import jax


OBJECT_TO_INDEX = {
	"unseen": 0,
	"empty": 1,
	"wall": 2,
	"onion": 3,
	"onion_pile": 4,
	"plate": 5,
	"plate_pile": 6,
	"goal": 7,
	"pot": 8,
	"dish": 9,
	"agent": 10,
}


COLORS = {
    'red'   : np.array([255, 0, 0]),
    'green' : np.array([0, 255, 0]),
    'blue'  : np.array([0, 0, 255]),
    'purple': np.array([112, 39, 195]),
    'yellow': np.array([255, 255, 0]),
    'grey'  : np.array([100, 100, 100]),
    'white' : np.array([255, 255, 255]),
	'black' : np.array([25, 25, 25]),
	'orange': np.array([230, 180, 0]),
}


COLOR_TO_INDEX = {
    'red'   : 0,
    'green' : 1,
    'blue'  : 2,
    'purple': 3,
    'yellow': 4,
    'grey'  : 5,
    'white' : 6,
	'black' : 7,
	'orange': 8,
}


OBJECT_INDEX_TO_VEC = jnp.array([
	jnp.array([OBJECT_TO_INDEX['unseen'], 0, 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['wall'], COLOR_TO_INDEX['grey'], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['onion'], COLOR_TO_INDEX["yellow"], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['onion_pile'], COLOR_TO_INDEX["yellow"], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['plate'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['plate_pile'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['goal'], COLOR_TO_INDEX['green'], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['pot'], COLOR_TO_INDEX['black'], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['dish'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8),
	jnp.array([OBJECT_TO_INDEX['agent'], COLOR_TO_INDEX['red'], 0], dtype=jnp.uint8),  					# Default color and direction
])


# Map of agent direction indices to vectors
DIR_TO_VEC = jnp.array([
	# Pointing right (positive X)
	(0, -1), # NORTH
	(0, 1), # SOUTH
	(1, 0), # EAST
	(-1, 0), # WEST
], dtype=jnp.int8)


def make_overcooked_map(
	wall_map,
	goal_pos,
	agent_pos,
	agent_dir_idx,
	plate_pile_pos,
	onion_pile_pos,
	pot_pos,
	pot_status,
	onion_pos,
	plate_pos,
	dish_pos,
	pad_obs=False,
	num_agents=2,
	agent_view_size=5):

	# Expand maze map to H x W x C
	empty = jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
	wall = jnp.array([OBJECT_TO_INDEX['wall'], COLOR_TO_INDEX['grey'], 0], dtype=jnp.uint8)
	maze_map = jnp.array(jnp.expand_dims(wall_map, -1), dtype=jnp.uint8)
	maze_map = jnp.where(maze_map > 0, wall, empty)

	# Add agents
	def _get_agent_updates(agent_dir_idx, agent_pos, agent_idx):
		agent = jnp.array([OBJECT_TO_INDEX['agent'], COLOR_TO_INDEX['red']+agent_idx*2, agent_dir_idx], dtype=jnp.uint8)
		agent_x, agent_y = agent_pos
		return agent_x, agent_y, agent

	agent_x_vec, agent_y_vec, agent_vec = jax.vmap(_get_agent_updates, in_axes=(0,0,0))(agent_dir_idx, agent_pos, jnp.arange(num_agents))
	maze_map = maze_map.at[agent_y_vec,agent_x_vec,:].set(agent_vec)

	# Add goals
	goal = jnp.array([OBJECT_TO_INDEX['goal'], COLOR_TO_INDEX['green'], 0], dtype=jnp.uint8)
	goal_x = goal_pos[:, 0]
	goal_y = goal_pos[:, 1]
	maze_map = maze_map.at[goal_y, goal_x, :].set(goal)

	# Add onions
	onion_x = onion_pile_pos[:,0]
	onion_y = onion_pile_pos[:,1]
	onion_pile = jnp.array([OBJECT_TO_INDEX['onion_pile'], COLOR_TO_INDEX["yellow"], 0], dtype=jnp.uint8)
	maze_map = maze_map.at[onion_y, onion_x, :].set(onion_pile)

	# Add plates
	plate_x = plate_pile_pos[:, 0]
	plate_y = plate_pile_pos[:, 1]
	plate_pile = jnp.array([OBJECT_TO_INDEX['plate_pile'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8)
	maze_map = maze_map.at[plate_y, plate_x, :].set(plate_pile)

	# Add pots
	pot_x = pot_pos[:, 0]
	pot_y = pot_pos[:, 1]
	pots = jnp.stack(
		[jnp.array([OBJECT_TO_INDEX['pot'], COLOR_TO_INDEX["black"], status]) for status in pot_status],
		dtype=jnp.uint8
	)
	maze_map = maze_map.at[pot_y, pot_x, :].set(pots)

	# TODO: maze_map += onion * onion_mask ...

	if len(onion_pos) > 0:
		onion_x = onion_pos[:, 0]
		onion_y = onion_pos[:, 1]
		onion = jnp.array([OBJECT_TO_INDEX['onion'], COLOR_TO_INDEX["yellow"], 0], dtype=jnp.uint8)
		maze_map = maze_map.at[onion_y, onion_x, :].set(onion)
	if len(plate_pos) > 0:
		plate_x = plate_pos[:, 0]
		plate_y = plate_pos[:, 1]
		plate = jnp.array([OBJECT_TO_INDEX['plate'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8)
		maze_map = maze_map.at[plate_y, plate_x, :].set(plate)
	if len(dish_pos) > 0:
		dish_x = dish_pos[:, 0]
		dish_y = dish_pos[:, 1]
		dish = jnp.array([OBJECT_TO_INDEX['dish'], COLOR_TO_INDEX["white"], 0], dtype=jnp.uint8)
		maze_map = maze_map.at[dish_y, dish_x, :].set(dish)

	# Add observation padding
	if pad_obs:
		padding = agent_view_size-1
	else:
		padding = 1

	maze_map_padded = jnp.tile(wall.reshape((1,1,*empty.shape)), (maze_map.shape[0]+2*padding, maze_map.shape[1]+2*padding, 1))
	maze_map_padded = maze_map_padded.at[padding:-padding,padding:-padding,:].set(maze_map)

	# Add surrounding walls
	wall_start = padding-1 # start index for walls
	wall_end_y = maze_map_padded.shape[0] - wall_start - 1
	wall_end_x = maze_map_padded.shape[1] - wall_start - 1
	maze_map_padded = maze_map_padded.at[wall_start,wall_start:wall_end_x+1,:].set(wall) # top
	maze_map_padded = maze_map_padded.at[wall_end_y,wall_start:wall_end_x+1,:].set(wall) # bottom
	maze_map_padded = maze_map_padded.at[wall_start:wall_end_y+1,wall_start,:].set(wall) # left
	maze_map_padded = maze_map_padded.at[wall_start:wall_end_y+1,wall_end_x,:].set(wall) # right

	return maze_map_padded



import chex
from absl import flags, app
import jax
import jax.numpy as jnp

from jaxmarl import make
from jaxmarl.environments.hanabi.hanabi import HanabiEnv

FLAGS = flags.FLAGS

flags.DEFINE_integer('num_players', default=2, help='Number of players playing the game.')
flags.DEFINE_integer('seed', default=0, help='Game seed. Used for deck generation.')

flags.register_validator(
    'num_players',
    lambda value: 2 <= value <= 5,
    message='--num_players must be between 2 and 5'
)


def batchify(x, agents):
    x = jnp.stack([x[a] for a in agents])
    return x.reshape((len(agents), -1))


def unbatchify(x, agents):
    return {a: x[i] for i, a in enumerate(agents)}


def get_user_action(player_legal_moves):
    while True:
        try:
            print("---")
            action = int(input("Pick Action: "))
            is_legal_action = player_legal_moves[action] == 1.0

            print(f"Is legal action: {is_legal_action}")
            if not is_legal_action:
                raise ValueError("Action is not legal.")

            print("---")
            return action
        except:
            print("Illegal action! Try again.")


def main(argv):
    num_players = FLAGS.num_players
    seed = FLAGS.seed

    print(f"Starting game with seed={seed}")
    key = jax.random.PRNGKey(seed)
    hand_size = 5 if num_players <= 3 else 4
    env: HanabiEnv = make("hanabi", hand_size=hand_size, num_agents=num_players)

    print(f"Action encoding for the environment: {env.action_encoding}")

    rng, _rng = jax.random.split(key)
    obs, env_state = env.reset(_rng)

    score = 0
    env_step_jit = jax.jit(env.step)
    while True:
        env.render(env_state)
        legal_moves_dict = env.get_legal_moves(env_state)
        legal_moves = batchify(legal_moves_dict, env.agents)

        cur_player = jnp.where(env_state.cur_player_idx == 1)[0][0]
        cur_player_legal_moves = legal_moves[cur_player]

        print("Legal moves for current player:")
        legal_moves_encoded = {
            i: env.action_encoding[i]
            for i, m
            in enumerate(cur_player_legal_moves)
            if m == 1.
        }
        print(legal_moves_encoded)

        user_action = get_user_action(cur_player_legal_moves)

        actions = jnp.full(
            env.num_agents, env.num_actions - 1
        ).at[cur_player].set(user_action)

        rng, _rng = jax.random.split(rng)
        actions = unbatchify(actions, env.agents)
        obs, env_state, rewards, dones, infos = env_step_jit(
            _rng, env_state, actions
        )
        print(f"Action played: {env.action_encoding[user_action]}")

        score += rewards['__all__']

        if dones['__all__']:
            break

    print(f"Game ended.\nScore: {score}")


if __name__ == '__main__':
    app.run(main)



""" Wrappers for use with jaxmarl baselines. """
import os
import jax
import jax.numpy as jnp
import chex
import numpy as np
from flax import struct
from functools import partial

# from gymnax.environments import environment, spaces
from gymnax.environments.spaces import Box as BoxGymnax, Discrete as DiscreteGymnax
from typing import Dict, Optional, List, Tuple, Union
from jaxmarl.environments.overcooked_v2.common import DynamicObject
from jaxmarl.environments.spaces import Box, Discrete, MultiDiscrete
from jaxmarl.environments.multi_agent_env import MultiAgentEnv, State

from safetensors.flax import save_file, load_file
from flax.traverse_util import flatten_dict, unflatten_dict

def save_params(params: Dict, filename: Union[str, os.PathLike]) -> None:
    flattened_dict = flatten_dict(params, sep=',')
    save_file(flattened_dict, filename)

def load_params(filename:Union[str, os.PathLike]) -> Dict:
    flattened_dict = load_file(filename)
    return unflatten_dict(flattened_dict, sep=",")


class JaxMARLWrapper(object):
    """Base class for all jaxmarl wrappers."""

    def __init__(self, env: MultiAgentEnv):
        self._env = env

    def __getattr__(self, name: str):
        return getattr(self._env, name)

    # def _batchify(self, x: dict):
    #     x = jnp.stack([x[a] for a in self._env.agents])
    #     return x.reshape((self._env.num_agents, -1))

    def _batchify_floats(self, x: dict):
        return jnp.stack([x[a] for a in self._env.agents])


@struct.dataclass
class LogEnvState:
    env_state: State
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int


class LogWrapper(JaxMARLWrapper):
    """Log the episode returns and lengths.
    NOTE for now for envs where agents terminate at the same time.
    """

    def __init__(self, env: MultiAgentEnv, replace_info: bool = False):
        super().__init__(env)
        self.replace_info = replace_info

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        obs, env_state = self._env.reset(key)
        state = LogEnvState(
            env_state,
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
        )
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: LogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        ep_done = done["__all__"]
        new_episode_return = state.episode_returns + self._batchify_floats(reward)
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info
    
@struct.dataclass
class OvercookedV2LogEnvState:
    env_state: State
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int
    returned_episode_recipe_returns: Dict[str, float]


class OvercookedV2LogWrapper(JaxMARLWrapper):
    def __init__(self, env: MultiAgentEnv, replace_info: bool = False):
        super().__init__(env)
        self.replace_info = replace_info

        self.recipe_dict = {
            f"{recipe[0]}_{recipe[1]}_{recipe[2]}": DynamicObject.get_recipe_encoding(
                recipe
            )
            for recipe in env.possible_recipes
        }

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        obs, env_state = self._env.reset(key)

        recipe_returns = {
            r: jnp.zeros((self._env.num_agents,)) for r in self.recipe_dict
        }

        state = OvercookedV2LogEnvState(
            env_state,
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            recipe_returns,
        )
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: OvercookedV2LogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        ep_done = done["__all__"]
        new_episode_return = state.episode_returns + self._batchify_floats(reward)
        new_episode_length = state.episode_lengths + 1

        updated_recipe_returns = {
            id: jax.lax.select(
                (state.env_state.recipe == self.recipe_dict[id]) & ep_done,
                new_episode_return,
                old_episode_return,
            )
            for id, old_episode_return in state.returned_episode_recipe_returns.items()
        }

        state = OvercookedV2LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=jax.lax.select(
                ep_done, new_episode_return, state.returned_episode_returns
            ),
            returned_episode_lengths=jax.lax.select(
                ep_done, new_episode_length, state.returned_episode_lengths
            ),
            returned_episode_recipe_returns=updated_recipe_returns,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        info["returned_episode_recipe_returns"] = state.returned_episode_recipe_returns
        return obs, state, reward, done, info


class MPELogWrapper(LogWrapper):
    """ Times reward signal by number of agents within the environment,
    to match the on-policy codebase. """
    
    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: LogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        rewardlog = jax.tree.map(lambda x: x*self._env.num_agents, reward)  # As per on-policy codebase
        ep_done = done["__all__"]
        new_episode_return = state.episode_returns + self._batchify_floats(rewardlog)
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info

@struct.dataclass
class SMAXLogEnvState:
    env_state: State
    episode_returns: float
    episode_lengths: int
    won_episode: int
    returned_episode_returns: float
    returned_episode_lengths: int
    returned_won_episode: int


class SMAXLogWrapper(JaxMARLWrapper):
    def __init__(self, env: MultiAgentEnv, replace_info: bool = False):
        super().__init__(env)
        self.replace_info = replace_info

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        obs, env_state = self._env.reset(key)
        state = SMAXLogEnvState(
            env_state,
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
            jnp.zeros((self._env.num_agents,)),
        )
        return obs, state

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: SMAXLogEnvState,
        action: Union[int, float],
    ) -> Tuple[chex.Array, LogEnvState, float, bool, dict]:
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action
        )
        ep_done = done["__all__"]
        batch_reward = self._batchify_floats(reward)
        new_episode_return = state.episode_returns + self._batchify_floats(reward)
        new_episode_length = state.episode_lengths + 1
        new_won_episode = (batch_reward >= 1.0).astype(jnp.float32)
        state = SMAXLogEnvState(
            env_state=env_state,
            won_episode=new_won_episode * (1 - ep_done),
            episode_returns=new_episode_return * (1 - ep_done),
            episode_lengths=new_episode_length * (1 - ep_done),
            returned_episode_returns=state.returned_episode_returns * (1 - ep_done)
            + new_episode_return * ep_done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - ep_done)
            + new_episode_length * ep_done,
            returned_won_episode=state.returned_won_episode * (1 - ep_done)
            + new_won_episode * ep_done,
        )
        if self.replace_info:
            info = {}
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_won_episode"] = state.returned_won_episode
        info["returned_episode"] = jnp.full((self._env.num_agents,), ep_done)
        return obs, state, reward, done, info


def get_space_dim(space):
    # get the proper action/obs space from Discrete-MultiDiscrete-Box spaces
    if isinstance(space, (DiscreteGymnax, Discrete)):
        return space.n
    elif isinstance(space, (BoxGymnax, Box, MultiDiscrete)):
        return np.prod(space.shape)
    else:
        print(space)
        raise NotImplementedError('Current wrapper works only with Discrete/MultiDiscrete/Box action and obs spaces')


class CTRolloutManager(JaxMARLWrapper):
    """
    Rollout Manager for Centralized Training of with Parameters Sharing. Used by JaxMARL Q-Learning Baselines.
    - Batchify multiple environments (the number of parallel envs is defined by batch_size in __init__).
    - Adds a global state (obs["__all__"]) and a global reward (rewards["__all__"]) in the env.step returns.
    - Pads the observations of the agents in order to have all the same length.
    - Adds an agent id (one hot encoded) to the observation vectors.

    By default:
    - global_state is the concatenation of all agents' observations.
    - global_reward is the sum of all agents' rewards.
    """

    def __init__(self, env: MultiAgentEnv, batch_size:int, training_agents:List=None, preprocess_obs:bool=True):
        
        super().__init__(env)
        
        self.batch_size = batch_size

        # the agents to train could differ from the total trainable agents in the env (f.i. if using pretrained agents)
        # it's important to know it in order to compute properly the default global rewards and state
        self.training_agents = self.agents if training_agents is None else training_agents  
        self.preprocess_obs = preprocess_obs  
        
        # batched action sampling
        self.batch_samplers = {agent: jax.jit(jax.vmap(self.action_space(agent).sample, in_axes=0)) for agent in self.agents}

        # assumes the observations are flattened vectors
        self.max_obs_length = max(list(map(lambda x: get_space_dim(x), self.observation_spaces.values())))
        self.max_action_space = max(list(map(lambda x: get_space_dim(x), self.action_spaces.values())))
        self.obs_size = self.max_obs_length
        if self.preprocess_obs:
            self.obs_size += len(self.agents)

        # agents ids
        self.agents_one_hot = {a:oh for a, oh in zip(self.agents, jnp.eye(len(self.agents)))}
        # valid actions
        self.valid_actions = {a:jnp.arange(u.n) for a, u in self.action_spaces.items()}
        self.valid_actions_oh ={a:jnp.concatenate((jnp.ones(u.n), jnp.zeros(self.max_action_space - u.n))) for a, u in self.action_spaces.items()}

        # custom global state and rewards for specific envs
        if 'smax' in env.name.lower():
            self.global_state = lambda obs, state: obs['world_state']
            self.global_reward = lambda rewards: rewards[self.training_agents[0]]
            self.get_valid_actions = lambda state: jax.vmap(env.get_avail_actions)(state)
        elif 'overcooked' in env.name.lower():
            self.global_state = lambda obs, state:  jnp.concatenate([obs[agent].flatten() for agent in self.agents], axis=-1)
            self.global_reward = lambda rewards: rewards[self.training_agents[0]]
        elif 'hanabi' in env.name.lower():
            self.global_reward = lambda rewards: rewards[self.training_agents[0]]
            self.get_valid_actions = lambda state: jax.vmap(env.get_legal_moves)(state)

    
    @partial(jax.jit, static_argnums=0)
    def batch_reset(self, key):
        keys = jax.random.split(key, self.batch_size)
        return jax.vmap(self.wrapped_reset, in_axes=0)(keys)

    @partial(jax.jit, static_argnums=0)
    def batch_step(self, key, states, actions):
        keys = jax.random.split(key, self.batch_size)
        return jax.vmap(self.wrapped_step, in_axes=(0, 0, 0))(keys, states, actions)

    @partial(jax.jit, static_argnums=0)
    def wrapped_reset(self, key):
        obs_, state = self._env.reset(key)
        if self.preprocess_obs:
            obs = jax.tree.map(self._preprocess_obs, {agent:obs_[agent] for agent in self.agents}, self.agents_one_hot)
        else:
            obs = obs_
        obs["__all__"] = self.global_state(obs_, state)
        return obs, state

    @partial(jax.jit, static_argnums=0)
    def wrapped_step(self, key, state, actions):
        obs_, state, reward, done, infos = self._env.step(key, state, actions)
        if self.preprocess_obs:
            obs = jax.tree.map(self._preprocess_obs, {agent:obs_[agent] for agent in self.agents}, self.agents_one_hot)
            obs = jax.tree.map(lambda d, o: jnp.where(d, 0., o), {agent:done[agent] for agent in self.agents}, obs) # ensure that the obs are 0s for done agents
        else:
            obs = obs_
        obs["__all__"] = self.global_state(obs_, state)
        reward["__all__"] = self.global_reward(reward)
        return obs, state, reward, done, infos

    @partial(jax.jit, static_argnums=0)
    def global_state(self, obs, state):
        return jnp.concatenate([obs[agent] for agent in self.agents], axis=-1)
    
    @partial(jax.jit, static_argnums=0)
    def global_reward(self, reward):
        return jnp.stack([reward[agent] for agent in self.training_agents]).sum(axis=0) 
    
    def batch_sample(self, key, agent):
        return self.batch_samplers[agent](jax.random.split(key, self.batch_size)).astype(int)
    
    @partial(jax.jit, static_argnums=0)
    def get_valid_actions(self, state):
        # default is to return the same valid actions one hot encoded for each env 
        return {agent:jnp.tile(actions, self.batch_size).reshape(self.batch_size, -1) for agent, actions in self.valid_actions_oh.items()}

    @partial(jax.jit, static_argnums=0)
    def _preprocess_obs(self, arr, extra_features):
        # flatten
        arr = arr.flatten()
        # pad the observation vectors to the maximum length
        pad_width = [(0, 0)] * (arr.ndim - 1) + [(0, max(0, self.max_obs_length - arr.shape[-1]))]
        arr = jnp.pad(arr, pad_width, mode='constant', constant_values=0)
        # concatenate the extra features
        arr = jnp.concatenate((arr, extra_features), axis=-1)
        return arr



import jax 
import jax.numpy as jnp
from functools import partial
import os
import pickle
from typing import Tuple, List
from .map import Map

import numpy as np
import chex
from enum import IntEnum
import matplotlib.axes._axes as axes

import jaxmarl.environments.jaxnav.jaxnav_graph_utils as _graph_utils

def rotation_matrix(theta: float) -> jnp.ndarray:
    """ Rotate about the z axis. Assume theta in radians """
    return jnp.array([
        [jnp.cos(theta), -jnp.sin(theta)],
        [jnp.sin(theta), jnp.cos(theta)]
    ])

class SampleTestCaseTypes(IntEnum):
    RANDOM = 0
    GRID = 1

class GridMapCircleAgents(Map):
    
    def __init__(self,
                 num_agents: int,
                 rad,
                 map_size: Tuple[int, int],
                 fill: float=0.4,
                 cell_size: float=1.0,
                 sample_test_case_type='random',
                 **map_kwargs):
        super().__init__(num_agents, rad, map_size, **map_kwargs)
        assert self.rad < 1  # collision code only works for radius <1
        
        self.width = map_size[0]
        self.height = map_size[1]
        self.length = self.width*self.height
        self.pos_offset = jnp.full((2,), 0.5)
        
        self.cell_size = cell_size
        self.scaled_rad = self.scale_coords(rad)
        self.circle_check_window = jnp.ceil(self.scaled_rad).astype(jnp.int32)
        idxs = jnp.arange(-self.circle_check_window-1, self.circle_check_window+1)
        self.cc_x_idx, self.cc_y_idx = jnp.meshgrid(idxs, idxs)
        self.cell_half_height = self.cell_size / 2
        
        # determine max number of blocks
        if sample_test_case_type == 'random':
            self.sample_test_case_type = SampleTestCaseTypes.RANDOM
        elif sample_test_case_type == 'grid':
            self.sample_test_case_type = SampleTestCaseTypes.GRID
        else:
            raise ValueError(f"Invalid sample_test_case_type: {sample_test_case_type}")
        self.fill = fill
        self.free_grids = (self.width-2)*(self.height-2)
        self.n_clutter = jnp.floor(self.free_grids*self.fill).astype(int)
    
    @partial(jax.jit, static_argnums=[0])
    def sample_test_case(self, rng: chex.PRNGKey):
        
        return jax.lax.switch(
            self.sample_test_case_type,
            [
                super().sample_test_case,
                self.grid_sample_test_case,
            ],
            rng
        )
        
    def grid_sample_test_case(self, key):
        """ NOTE this won't throw an error if it's not possible, will just loop forever"""
        assert self.cell_size == 1.0
        
        key, _key = jax.random.split(key)
        map_data = self.sample_map(_key)
        inside_grid = map_data.at[1:-1, 1:-1].get()
        iwidth = self.width - 2
        
        def _sample_pair(key, start_masks, goal_masks):
            
            flat_occ = start_masks.flatten()
            key, _key = jax.random.split(key)
            start_idx = jax.random.choice(_key, len(flat_occ), (1,), p=(1-flat_occ))[0]
            start = jnp.array([start_idx % iwidth, start_idx // iwidth])  # [x, y]
            actual_idx = (start + 1).astype(jnp.int32)
            # connected_region = _graph_utils.component_mask_with_pos(inside_grid, start_idx)  # BUG not working on inside grid
            if self.valid_path_check:
                connected_region = _graph_utils.component_mask_with_pos(map_data, actual_idx).at[1:-1, 1:-1].get()
            else:
                connected_region = 1-inside_grid
            masked_start = connected_region.at[start[1], start[0]].set(0)
            goal_possibilities = masked_start & (1 - goal_masks)
            valid = jnp.any(goal_possibilities)  # only valid if possible goal locations
                        
            goal_idx = jax.random.choice(key, len(flat_occ), (1,), p=goal_possibilities.flatten())[0]
            goal = jnp.array([goal_idx % iwidth, goal_idx // iwidth])  # [x, y]
            
            return valid, start, goal
            
        def scan_fn(carry, rng):
            i, pos, start_mask, goal_mask = carry
            def _cond_fn(val):                
                return val[0] 
            
            def _body_fn(val):
                valid, rng, pos = val
                
                rng, _rng_pair = jax.random.split(rng)
                valid, start, goal = _sample_pair(_rng_pair, start_mask, goal_mask)
                
                positions = jnp.concatenate([start[None], goal[None]], axis=0)
                pos = pos.at[i].set(positions)
                
                return jnp.bitwise_not(valid), rng, pos
            
            (_, rng, pos) = jax.lax.while_loop(
                _cond_fn,
                _body_fn,
                (True, rng, pos)                
            )
            
            start = pos.at[i, 0].get().astype(jnp.int32)
            goal = pos.at[i, 1].get().astype(jnp.int32)
            start_mask = start_mask.at[start[1], start[0]].set(1)
            goal_mask = goal_mask.at[goal[1], goal[0]].set(1)
            return (i+1, pos, start_mask, goal_mask), None
        
                
        fill_max = jnp.max(jnp.array(self.map_size)) + self.rad*2
        pos = jnp.full((self.num_agents, 2, 2), fill_max)  # [num_agents, [start_pose, goal_pose]]
        
        key, key_scan = jax.random.split(key)
        key_scan = jax.random.split(key_scan, self.num_agents)
        (_, pos, _, _), _ = jax.lax.scan(scan_fn, (0, pos, inside_grid, inside_grid), key_scan)
        theta = jax.random.uniform(key, (self.num_agents, 2, 1), minval=-jnp.pi, maxval=jnp.pi)
        cases = jnp.concatenate([pos + 1.5, theta], axis=2)
        
        return map_data, cases
        
    @partial(jax.jit, static_argnums=[0])
    def sample_map(self, key):
        """ Sample map grid from uniform distribution """

        key_fill, key_shuff = jax.random.split(key)

        base_map = self._gen_base_grid()

        free_idx = jnp.arange(0, self.free_grids)
        num_fill = jax.random.randint(key_fill, (1,), 0, self.n_clutter)[0]
        
        map_within = jnp.where(free_idx<num_fill, 1, 0)
        #map_fill = jax.random.shuffle(key_shuff, map_within)
        map_fill = jax.random.permutation(key_shuff, map_within, independent=True)
        map_fill = jnp.reshape(map_fill, (self.height-2, self.width-2))
        return base_map.at[1:-1, 1:-1].set(map_fill)

    
    ### === Collision checking === ###
    @partial(jax.jit, static_argnums=[0])
    def check_circle_map_collision(self, pos, map_grid, rad=None):
        """ For a circle agent"""
        def _variable_grid_size_check(pos, rad, map_grid):
            
            # check inside
            @partial(jax.vmap, in_axes=(0, 0))
            def __check_inside(x_idx, y_idx):
                rc = jnp.array([x_idx, y_idx]) + 0.5
                dist_l2 = jnp.linalg.norm(scaled_pos - rc)
                c = jnp.minimum(dist_l2 - self.scaled_rad, 0.0)
                return c <= 0
            scaled_pos = self.scale_coords(pos)
            
            grid_pos = jnp.floor(scaled_pos).astype(jnp.int32)
            x_idx = self.cc_x_idx + grid_pos[0]
            y_idx = self.cc_y_idx + grid_pos[1]
            
            c_inside = __check_inside(x_idx.flatten(), y_idx.flatten())
            col_check = jnp.any(map_grid.at[y_idx.flatten(), x_idx.flatten()].get() & c_inside)
            
            # check perimeter
            theta = jnp.linspace(0, 2*jnp.pi, 100)  # NOTE hardcoded
            x_cir = rad * jnp.cos(theta) + pos[0] 
            y_cir = rad * jnp.sin(theta) + pos[1] 
            
            x_cir_idx = jnp.floor(self.scale_coords(x_cir)).astype(jnp.int32) 
            y_cir_idx = jnp.floor(self.scale_coords(y_cir)).astype(jnp.int32) 
            col_check = jnp.any(map_grid.at[x_cir_idx, y_cir_idx].get()) | col_check
            
            return col_check
        
        def _grid_size_of_1_check(pos, rad, map_grid):
            h, w = map_grid.shape 
            min_x, min_y = jnp.floor(jnp.maximum(jnp.zeros(2), pos-rad)).astype(int)
            max_x, max_y = jnp.floor(jnp.minimum(jnp.array([w, h]), pos+rad)).astype(int)
                    
            map_c_list = jnp.array([
                [min_x, min_y],
                [max_x, min_y],
                [min_x, max_y],
                [max_x, max_y],
            ]) + self.cell_half_height
                
            grid_check = self._check_grid(map_c_list, pos, rad)
            
            map_occ = jnp.array([
                map_grid[min_y, min_x],
                map_grid[min_y, max_x],
                map_grid[max_y, min_x],
                map_grid[max_y, max_x],
            ]).astype(int)
            return jnp.any((map_occ+grid_check)>1)
        if rad is None: rad = self.rad
        
        return jax.lax.switch(
            int(self.cell_size == 1.0),
            [
                _variable_grid_size_check,
                _grid_size_of_1_check,
            ],
            *(pos, rad, map_grid)
        )    
    
    
    def get_circle_map_occupancy_mask(self, pos, map_grid, rad=None):
        if rad is None: rad=self.rad
        
        wall_map = jnp.zeros(map_grid.shape, dtype=jnp.int32)
        
        theta = jnp.linspace(0, 2*jnp.pi, 100)
        pos_to_check = jnp.array([jnp.cos(theta), jnp.sin(theta)]).T * rad + pos
        print('pos to check shape', pos_to_check.shape)
        idxs = jnp.floor(pos_to_check).astype(int)
        wall_map = wall_map.at[idxs[:, 1], idxs[:, 0]].set(1)
        
        x_mesh, y_mesh = jnp.meshgrid(jnp.arange(0, self.width), jnp.arange(0, self.height))
        mesh = jnp.dstack([x_mesh, y_mesh]).reshape((-1,2))
        cc = jnp.linalg.norm(mesh - pos, axis=1) < rad
        inside_mask = cc.reshape((self.height, self.width)).astype(int)
        print('inside mask', inside_mask)
        return wall_map | inside_mask
        
    
    def check_agent_beam_intersect(self, beam, pos, theta, range_resolution, rad=None):
        """ Check for intersection between a lidar beam and an agent. """
        if rad is None: rad = self.rad
        d = beam[-1] - beam[0]
        f = beam[0] - pos
        
        a = jnp.dot(d, d)
        b = 2*jnp.dot(f, d)
        c = jnp.dot(f, f) - self.rad**2
        
        descrim = b**2 - 4*a*c
        
        t1 = (-b - jnp.sqrt(descrim))/(2*a)
        # t2 = (-b + jnp.sqrt(descrim))/(2*a)
        
        miss = (descrim < 0) | (t1 < 0) | (t1 > 1) #| (host_idx==other_idx)  # | (t2 < 0) | (t2 > 1)
        
        intersect = beam[0] + t1*d
        idx = jnp.floor(jnp.linalg.norm(intersect - beam[0])/range_resolution).astype(int)
        return jax.lax.select(miss, -1, idx)
    
    @partial(jax.jit, static_argnums=[0])
    def check_point_map_collision(self, pos, map_grid):
        """ For a point """
        pos = jnp.floor(self.scale_coords(pos)).astype(int)
        return map_grid.at[pos[1], pos[0]].get() == 1
    
    def check_all_agent_agent_collisions(self, agent_positions: chex.Array, agent_theta: chex.Array) -> chex.Array:
        
        @partial(jax.vmap, in_axes=(0, None))
        def _check_agent_collisions(agent_idx: int, agent_positions: chex.Array) -> bool:
            # TODO this function is a little clunky FIX 
            z = jnp.zeros(agent_positions.shape)
            z = z.at[agent_idx,:].set(jnp.ones(2)*self.rad*2.1)  
            x = agent_positions + z
            return jnp.any(jnp.sqrt(jnp.sum((x - agent_positions[agent_idx,:])**2, axis=1)) <= self.rad*2) 
        
        return _check_agent_collisions(jnp.arange(agent_positions.shape[0]), agent_positions)
    
    @partial(jax.jit, static_argnums=[0])
    def _gen_base_grid(self):
        """ Generate base grid map with walls around border """
        
        map = jnp.zeros((self.height, self.width), dtype=int)
        map = map.at[0,:].set(1)
        map = map.at[-1,:].set(1)
        map = map.at[:, 0].set(1)
        map = map.at[:, -1].set(1)
        
        return map
    
    @partial(jax.vmap, in_axes=[None, 0, None, None])
    def _check_grid(self, c, pos, radius):
        p = jnp.clip(pos - c, -self.cell_half_height, self.cell_half_height) 
        p = p + c 
        return jnp.linalg.norm(p - pos) <= radius
    
    @partial(jax.jit, static_argnums=[0, 4])
    def check_line_collision(self, pos1, pos2, map_data, max_length=5.0):
        """ uses same method as lidar (ray tracing) """
        resolution = 0.05
        line_length = jnp.linalg.norm(pos2-pos1)
        angle = jnp.arctan2(pos2[1]-pos1[1], pos2[0]-pos1[0])

        points = jnp.arange(0, max_length, resolution)
        points = points[:, None] * jnp.array([jnp.cos(angle), jnp.sin(angle)]) + pos1
        coords = jnp.floor(points).astype(int)
        lidar_hits = map_data[coords[:, 1], coords[:, 0]]

        num_points = jnp.floor(line_length/resolution).astype(int)
        idx_range = jnp.arange(points.shape[0])
        lidar_mask = jnp.where(idx_range<num_points, 1, 0)
        lidar_hits = lidar_hits * lidar_mask

        return jnp.any(lidar_hits)
    
    @partial(jax.jit, static_argnums=[0])
    def passable_check(self, pos1, pos2, map_data):
        """ Check if a path exists between pos1 and pos2"""
        
        def _passable(grid, posa, posb):
            grid = _graph_utils.component_mask_with_pos(grid, posa)
            return grid.at[posb[1], posb[0]].get()
        
        grid = map_data.astype(jnp.bool_)
        pos1 = jnp.floor(pos1).astype(jnp.int32)
        pos2 = jnp.floor(pos2).astype(jnp.int32)
        
        if len(pos2.shape) == 2: # batch eval 
            return jax.vmap(_passable, in_axes=(None, 0, 0))(
                grid, pos1, pos2
            )
        else:
            return _passable(grid, pos1, pos2)
        
    def scale_coords(self, x):
        return x / self.cell_size
    
    def dikstra_path(self, map_data, pos1, pos2):
        """ Computes shorted path (if possible) between `pos1` and `pos2` on the grid specified by `map_data`.
        
        Method TL;DR: Dijkstra's algorithm. JIT'd while loop through the open set of nodes, updating the distance
                      to go for each neighbour in `_body`. Terminates when the open set is empty.
        
        TODO: also terminate if the goal is reached. Add unit tests.
        
        Output:
        - `valid` (bool): True if a path exists
        - `d` (float): distance of shortest path, INF if no path exists
        """
        
        h, w = map_data.shape
        NEIGHBOURS = jnp.array([[-1, 0], [1, 0], [0, -1], [0, 1]])  # left, right, up, down
        INF = 1e8  # large number to represent infinity (jnp.inf produced unwanted nan's)
        
        def _flatten_idx(i, j):
            return i * w + j
        
        def _unflatten_idx(idx):
            return jnp.array([idx // w, idx % w])
        
        start = jnp.floor(pos1).astype(jnp.int32)
        end = jnp.floor(pos2).astype(jnp.int32)
        
        h, w = map_data.shape
        start_flat_idx = start[0] * w + start[1]
        set_distances_to_go = jnp.ones((h*w,)) * INF
        
        set_distances_to_go = set_distances_to_go.at[start_flat_idx].set(0)
        set_visited = jnp.zeros((h*w,)) 
        flat_node_idx = jnp.argmin(set_distances_to_go + INF * set_visited)
        
        carry = (
            flat_node_idx,
            set_distances_to_go,
            set_visited
        )
        
        def _cond(carry):
            return carry[1][carry[0]] < INF 
        
        def _body(carry):
            flat_node_idx, set_distances_to_go, set_visited = carry
            unflat_node_idx = _unflatten_idx(flat_node_idx)
            n_idx = unflat_node_idx + NEIGHBOURS
            flat_n_idx = jax.vmap(_flatten_idx)(n_idx[:, 0], n_idx[:, 1])
            
            n_visited = set_visited.at[flat_n_idx].get()
            n_distances_to_go = set_distances_to_go.at[flat_n_idx].get()
            new_distance = set_distances_to_go[flat_node_idx] + 1
            n_valid_node = map_data[n_idx[:, 1], n_idx[:, 0]] == 0
            updated_neighbour_distances = jnp.where(
                (n_distances_to_go > new_distance) &
                (n_visited == 0) & 
                (n_valid_node), 
                new_distance, 
                n_distances_to_go)
            
            set_distances_to_go = set_distances_to_go.at[flat_n_idx].set(updated_neighbour_distances)
            set_visited = set_visited.at[flat_node_idx].set(1)
            flat_node_idx = jnp.argmin(set_distances_to_go + INF * set_visited)
            return (flat_node_idx, set_distances_to_go, set_visited)
            
        _, set_distances_to_go, _ = jax.lax.while_loop(
            _cond,
            _body,
            carry
        )
        d = set_distances_to_go[_flatten_idx(end[0], end[1])]
        valid = d < INF
        return valid, jax.lax.select(valid, d, 0.0)
        
    ### === VISUALIZATION === ###
    def plot_map(self, ax: axes.Axes, map_grid: jnp.ndarray):
        """ Plot map grid and scale xticks """
        # print('plotting map of size ', map_grid.shape, 'with cell size', self.cell_size)
        ax.imshow(map_grid, extent=(0, map_grid.shape[1], 0, map_grid.shape[0]), origin="lower", cmap='binary', zorder=0)
        ax.set_xlim(0.0, self.map_size[0])
        ax.set_ylim(0.0, self.map_size[1])
        
        # set ticks to be scaled for grid size
        step_size = 1 / self.cell_size
        x_range = jnp.arange(0, map_grid.shape[1]+1, step_size)
        y_range = jnp.arange(0, map_grid.shape[0]+1, step_size)
        ax.set_xticks(x_range)
        ax.set_xticklabels(x_range*self.cell_size)
        ax.set_yticks(y_range)
        ax.set_yticklabels(y_range*self.cell_size)
        
    # def plot_pose(self, ax: axes.Axes, pose):
    #     ax.scatter(pose[:, 0, 0], pose[:, 0, 1], c='b', marker='+')
    #     ax.scatter(pose[:, 1, 0], pose[:, 1, 1], c='r', marker='+')
        
    def plot_agents(self,
                    ax: axes.Axes,
                    pos: jnp.ndarray,
                    theta: jnp.ndarray,
                    goal: jnp.ndarray,
                    done: jnp.ndarray,
                    plot_line_to_goal=True,
                    colour_agents_by_idx=False,
                    rad=None):
        """ Plot agents """
        from matplotlib.patches import Circle
        if rad is None: rad = self.rad
        colors = ['red', 'blue', 'green', 'purple', 'orange', 'pink', 'yellow', 'brown', 'grey', 'cyan']
        if (done.shape[0] > len(colors)) and colour_agents_by_idx:
            print('Too many agents to colour by index')
            colour_agents_by_idx = False
            
        colours = ['black' if done else 'red' for done in done]
        if colour_agents_by_idx:
            colours = ['black' if done else colors[i] for i, done in enumerate(done)]

        for i in range(done.shape[0]):
            circle = Circle(pos[i], rad, facecolor=colours[i])
            ax.add_patch(circle)

            x = pos[i][0] + rad * np.cos(theta[i])
            y = pos[i][1] + rad * np.sin(theta[i])
            ax.plot([pos[i][0], x], [pos[i][1], y], color='black')
            
            if plot_line_to_goal:
                ax.plot([pos[i][0], goal[i][0]], [pos[i][1], goal[i][1]], color='black', alpha=0.5)
  
    def plot_agent_path(self,
                        ax: axes.Axes,
                        x_seq: jnp.ndarray,
                        y_seq: jnp.ndarray,):
        """ Plot agent path """
        x = self.scale_coords(x_seq)
        y = self.scale_coords(y_seq)
        ax.plot(x, y, c='b', linewidth=2.0, zorder=1)
  
default_coords = jnp.array([
    [-0.25, -0.25],
    [-0.25, 0.25],
    [0.25, 0.25],
    [0.25, -0.25],
])
jackal_coords = jnp.array([
    [-0.254, -0.215],
    [-0.254, 0.215],
    [0.254, 0.215],
    [0.254, -0.215],
])
middle_line = jnp.array([
    [0.0, 0.0],
    [0.254, 0.0],
])

class GridMapPolygonAgents(GridMapCircleAgents):
    """ Map for homogenous, convex polygon agents. 
    """
    
    def __init__(self,
                 num_agents: int,
                 rad,
                 map_size,
                 agent_coords=default_coords,
                 middle_line=middle_line,
                 **map_kwargs):
        super().__init__(num_agents, rad, map_size, **map_kwargs)
        
        self.agent_coords=agent_coords
        self.middle_line=middle_line
        
        # define window around agent to check for map collisions
        min_x = jnp.min(agent_coords[:, 0])
        max_x = jnp.max(agent_coords[:, 0])
        min_y = jnp.min(agent_coords[:, 1])
        max_y = jnp.max(agent_coords[:, 1])
        side_length = jnp.maximum(max_x - min_x, max_y - min_y)
                
        cell_ratio = side_length / self.cell_size
        
        self.agent_window = jnp.ceil(cell_ratio*2).astype(int)  # NOTE times 2 should be enough
        self.idx_offsets = jnp.arange(-self.agent_window, self.agent_window+1)        
        
        #  2D with one set of coords for all agents 
        assert (len(self.agent_coords.shape) == 2)
        # or \
            # ((self.agent_coords.shape[0] == self.num_agents) and \
                # (len(self.agent_coords.shape) == 3))

    @partial(jax.jit, static_argnums=[0])
    def transform_coords(self, pos, theta, coords):
        r = rotation_matrix(theta)
        return jnp.matmul(r, coords.T).T + pos

    @partial(jax.jit, static_argnums=[0])
    def check_agent_map_collision(self, pos, theta, map_grid, agent_coords=None):
        """ Check for collision between agent and map. For polygon agents. 
        For now assuming all agents have the same shape and that side lengths 
        are less than the grid size. """   
                
        if agent_coords is None: agent_coords = self.agent_coords
                
        idx_to_check = jnp.floor(pos / self.cell_size).squeeze()  # [x, y]
        idx_0 = (idx_to_check[0] + self.idx_offsets).astype(int)
        idx_1 = (idx_to_check[1] + self.idx_offsets).astype(int)

        idx_pairs = jax.vmap(
            lambda x, y: jax.vmap(lambda a, b: jnp.array([a, b]), in_axes=(None, 0))(x, y),
            in_axes=(0, None)
        )(idx_1, idx_0).reshape((-1, 2))
        
        # need to scale to take account of grid size
        transformed_coords = self.transform_coords(pos, theta.squeeze(), agent_coords)
        scaled_coords = transformed_coords / self.cell_size
        scaled_coords_appended = jnp.concatenate([scaled_coords, scaled_coords[0, :][None]], axis=0)

        cut = jnp.any(
            jax.vmap(self._checkGrid,
                     in_axes=(None, None, 0, None))(scaled_coords_appended[:-1], scaled_coords_appended[1:], idx_pairs, map_grid)
        )
        
        inside = jnp.any(
            jax.vmap(self._checkInsideGrid,
                     in_axes=(None, 0, None))(scaled_coords, idx_pairs, map_grid)
        )
        return cut | inside
        
    @partial(jax.jit, static_argnums=[0])
    def get_agent_map_occupancy_mask(self, pos, theta, map_grid, agent_coords=None):
        
        if agent_coords is None: agent_coords = self.agent_coords
        
        map_mask = jnp.ones(map_grid.shape, dtype=jnp.int32)
        
        idx_to_check = jnp.floor(pos / self.cell_size).squeeze()
        idx_0 = (idx_to_check[0] + self.idx_offsets).astype(int)
        idx_1 = (idx_to_check[1] + self.idx_offsets).astype(int)

        idx_pairs = jax.vmap(
            lambda x, y: jax.vmap(lambda a, b: jnp.array([a, b]), in_axes=(None, 0))(x, y),
            in_axes=(0, None)
        )(idx_1, idx_0).reshape((-1, 2))
        
        # need to scale to take account of grid size
        transformed_coords = self.transform_coords(pos, theta.squeeze(), agent_coords)
        scaled_coords = transformed_coords / self.cell_size
        scaled_coords_appended = jnp.concatenate([scaled_coords, scaled_coords[0, :][None]], axis=0)

        cut = jax.vmap(self._checkGrid,
                       in_axes=(None, None, 0, None))(scaled_coords_appended[:-1], scaled_coords_appended[1:], idx_pairs, map_mask)
        inside = jax.vmap(self._checkInsideGrid,
                          in_axes=(None, 0, None))(scaled_coords, idx_pairs, map_mask)
        
        collisions = cut | inside
        valid_idx = (idx_pairs[:, 0] >= 0) & (idx_pairs[:, 0] < self.height) & (idx_pairs[:, 1] >= 0) & (idx_pairs[:, 1] < self.width) & collisions
        idx_pairs = jnp.where(jnp.repeat(valid_idx[:, None], 1, 1), idx_pairs, jnp.zeros(idx_pairs.shape)).astype(int)
        map_mask = jnp.zeros(map_grid.shape, dtype=jnp.int32)
        
        return map_mask.at[idx_pairs[:, 0], idx_pairs[:, 1]].set(1)

    
    def _checkGrid(self, x1y1, x2y2, grid_idx, map_grid):
        
        def _checkLineLine(x1, y1, x2, y2, x3, y3, x4, y4):
            """ Check collision between line (x1, y1) -- (x2, y2) and line (x3, y3) -- (x4, y4) """
            uA = ((x4-x3)*(y1-y3) - (y4-y3)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
            uB = ((x2-x1)*(y1-y3) - (y2-y1)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
            c = (uA >= 0) & (uA <= 1) & (uB >= 0) & (uB <= 1)
            return c.astype(jnp.bool_)      
        
        def _checkLineRect(x1, y1, x2, y2, rx, ry):
            """ Check collision between line (x1, y1) -- (x2, y2) and rectangle with bottom left corner at (rx, ry) 
            and width and height of 1."""
            vmap_checkLineLine = jax.vmap(_checkLineLine, in_axes=(None, None, None, None, 0, 0, 0, 0))
            x3 = jnp.array([0, 1, 0, 0]) + rx
            y3 = jnp.array([0, 0, 0, 1]) + ry
            x4 = jnp.array([0, 1, 1, 1]) + rx
            y4 = jnp.array([1, 1, 0, 1]) + ry
            checks = vmap_checkLineLine(x1, y1, x2, y2, x3, y3, x4, y4)
            return jnp.any(checks)
        
        @partial(jax.vmap, in_axes=(0, 0, None))
        def _checkSide(x1y1, x2y2, grid_idx):
            x1, y1 = x1y1
            x2, y2 = x2y2
            ry, rx = grid_idx[0], grid_idx[1]
            filled = map_grid[ry, rx]
            c = _checkLineRect(x1, y1, x2, y2, rx, ry)
            return c & filled
        
        valid_idx = (grid_idx[0] >= 0) & (grid_idx[0] < self.height) & (grid_idx[1] >= 0) & (grid_idx[1] < self.width)
        return jnp.any(_checkSide(x1y1, x2y2, grid_idx)) & valid_idx
    
    def _checkInsideGrid(self, sides, grid_idx, map_grid):
        """ Check if polygon is inside grid cell, NOTE assumes grid cell is of size 1x1 """
        
        def _checkPolyWithinRect(sides, rx, ry):
            """ Check if polygon is within rectangle with bottom left corner at (rx, ry) and width and height of 1."""
            
            def _checkPointRect(x, y, rx, ry):
                """ Check if point (x, y) is within rectangle with bottom left corner at (rx, ry) and width and height of 1."""
                return (x >= rx) & (x <= rx+1) & (y >= ry) & (y <= ry+1)
            
            vmap_checkPointRect = jax.vmap(_checkPointRect, in_axes=(0, 0, None, None))
            checks = vmap_checkPointRect(sides[:, 0], sides[:, 1], rx, ry)
            return jnp.all(checks)
        
        valid_idx = (grid_idx[0] >= 0) & (grid_idx[0] < self.height) & (grid_idx[1] >= 0) & (grid_idx[1] < self.width)
        inside = _checkPolyWithinRect(sides, grid_idx[1], grid_idx[0])
        return inside & map_grid[grid_idx[0], grid_idx[1]] & valid_idx
    
    @partial(jax.jit, static_argnums=[0])
    def check_all_agent_agent_collisions(self, agent_positions: chex.Array, agent_theta: chex.Array, agent_coords=None) -> chex.Array:
        """ Use Separating Axis Theorem (SAT) to check for collisions between convex polygon agents. 
        
        Separating Axis Theorem TL;DR: Searches for a line that separates two convex polygons. If no line is found, the polygons are colliding.
                                       To search for a separating line, we look at the normals of the edges of the polygons and project the vertices
                                       of all polygons onto these normals. If the projections of the vertices of one polygon do not overlap with the
                                       projections of the vertices of the other polygon, then the polygons are not colliding.
        """    
        
        def _orthogonal(v):
            return jnp.array([v[1], -v[0]])
        
        if agent_coords is None: agent_coords = self.agent_coords
        
        transformed_coords = jax.vmap(self.transform_coords, in_axes=(0, 0, None))(agent_positions, agent_theta.squeeze(), agent_coords)
        all_coords = transformed_coords.reshape((-1, 2))  # [num_agents*4, 2]
        trans_rolled = jnp.roll(transformed_coords, 1, axis=1)
        
        edges = transformed_coords - trans_rolled  # [num_agents, 4, 2]
        orthog_edges = jax.vmap(_orthogonal, in_axes=(0))(edges.reshape((-1, 2))).reshape((-1, 4, 2))  # NOTE assuming 4 sides

        all_axis = orthog_edges / jnp.linalg.norm(orthog_edges, axis=2)[:,:,None]  # [num_agents, 4, 2]
            
        def _project_axis(axis):
            return jax.vmap(jnp.dot, in_axes=(None, 0))(axis, all_coords)
            
        axis_projections = jax.vmap(_project_axis)(all_axis.reshape((-1, 2)))
        axis_projections_by_agent = axis_projections.reshape((-1, self.num_agents, 4)) 
        axis_ranges_by_agent = jnp.stack([jnp.min(axis_projections_by_agent, axis=2), jnp.max(axis_projections_by_agent, axis=2)], axis=-1)
        axis_by_agent_range_by_agent = axis_ranges_by_agent.reshape((self.num_agents, -1) + axis_ranges_by_agent.shape[1:])
        def _calc_overlaps(agent_idx, agent_axis_ranges):
            overlaps = (agent_axis_ranges[:, agent_idx, 0][:, None] <= agent_axis_ranges[:, :, 1]) & (agent_axis_ranges[:, :, 0] <= agent_axis_ranges[:, agent_idx, 1][:, None])            
            overlaps = overlaps.at[:, agent_idx].set(False)
            return jnp.all(overlaps, axis=0)
        
        overlaps_matrix = jax.vmap(_calc_overlaps)(jnp.arange(self.num_agents), axis_by_agent_range_by_agent) # all overlaps for all vertex is a collision
        # need to match matrix triangles
        def _join_matrix(rows, cols):
            return jnp.any(jnp.bitwise_and(rows, cols))
            
        c_free = jax.vmap(_join_matrix)(overlaps_matrix, overlaps_matrix.T)
        return c_free
                
    def check_agent_beam_intersect(self, beam, pos, theta, range_resolution, agent_coords=None):
        """ Check for intersection between a lidar beam and an agent. """
        if agent_coords is None: agent_coords = self.agent_coords
        
        @partial(jax.vmap, in_axes=(None, None, 0, 0))
        def _checkSide(beam_start, beam_end, side_start, side_end):
            """ Check collision between line (x1, y1) -- (x2, y2) and line (x3, y3) -- (x4, y4) """
            x1, y1 = beam_start
            x2, y2 = beam_end
            x3, y3 = side_start
            x4, y4 = side_end
            
            uA = ((x4-x3)*(y1-y3) - (y4-y3)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
            uB = ((x2-x1)*(y1-y3) - (y2-y1)*(x1-x3)) / ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1))
            c = (uA >= 0) & (uA <= 1) & (uB >= 0) & (uB <= 1)
            
            ix = x1 + (uA * (x2-x1))
            iy = y1 + (uA * (y2-y1))
            intersect = jnp.array([ix, iy])
            idx = jnp.floor(jnp.linalg.norm(intersect - beam[0])/range_resolution)
            
            return jax.lax.select(c, idx, jnp.inf)
        
        tc = self.transform_coords(pos, theta, agent_coords)
        tc = jnp.concatenate([tc, tc[0, :][None]], axis=0)
        
        idxs = _checkSide(beam[0], beam[-1], tc[:-1], tc[1:])
        idx = jnp.min(idxs)
        return jax.lax.select(idx==jnp.inf, -1.0, idx).astype(int)

    def plot_agents(
                self,
                ax: axes.Axes,
                pos: jnp.ndarray,
                theta: jnp.ndarray,
                goal: jnp.ndarray,
                done: jnp.ndarray,
                plot_line_to_goal=True,
                agent_coords=None,
                middle_line=None,
                colour_agents_by_idx=False,
        ):
        """ Plot agents """
        from matplotlib.patches import Polygon
        if agent_coords is None: agent_coords = self.agent_coords
        if middle_line is None: middle_line = self.middle_line
        colors = ['red', 'blue', 'green', 'purple', 'orange', 'pink', 'yellow', 'brown', 'grey', 'cyan']
        if (done.shape[0] > len(colors)) and colour_agents_by_idx:
            print('Too many agents to colour by index')
            colour_agents_by_idx = False
            
        colours = ['black' if done else 'red' for done in done]
        if colour_agents_by_idx:
            colours = ['black' if done else colors[i] for i, done in enumerate(done)]

        for i in range(done.shape[0]):
            transformed_coords = self.transform_coords(pos[i], theta[i], agent_coords) / self.cell_size
            
            poly = Polygon(transformed_coords, facecolor=colours[i])
            ax.add_patch(poly)

            # middle line
            transformed_middle_line = self.transform_coords(pos[i], theta[i], self.middle_line) / self.cell_size
            ax.plot(transformed_middle_line[:, 0], transformed_middle_line[:, 1], color='black')
                        
            if plot_line_to_goal:
                pos_scaled = self.scale_coords(pos[i])
                goal_scaled = self.scale_coords(goal[i])
                ax.scatter(goal_scaled[0], goal_scaled[1], marker='+', c='g')
                ax.plot([pos_scaled[0], goal_scaled[0]], [pos_scaled[1], goal_scaled[1]], color='black', alpha=0.5) 

SMOOTHING_IDX_OFFSETS = jnp.array([[0, 1], [0, -1], [1, 0], [-1, 0], [1, 1], [1, -1], [-1, 1], [-1, -1]])
INFLATE_IDX_OFFSETS_3 = SMOOTHING_IDX_OFFSETS
INFLATE_IDX_OFFSETS_5 = jnp.concatenate([INFLATE_IDX_OFFSETS_3, jnp.array([[2, 0], [2, 1], [2, 2], [1, 2], [0, 2], [-1, 2], [-2, 2], [-2, 1], [-2, 0], [-2, -1], [-2, -2], [-1, -2], [0, -2], [1, -2], [2, -2], [2, -1],])], axis=0)


class GridMapBarn(GridMapPolygonAgents):
    
    def __init__(self,
                 num_agents, 
                 rad, 
                 map_size,
                 smoothing_iters=4,
                 smoothing_upper_count=3,
                 smoothing_lower_count=1,
                 agent_coords=jackal_coords,
                 cell_size=0.15,
                 **map_kwargs):
        
        super().__init__(num_agents, rad, map_size, agent_coords=agent_coords, cell_size=cell_size, **map_kwargs)

        self.smoothing_iters = smoothing_iters
        self.smoothing_upper_count = smoothing_upper_count
        self.smoothing_lower_count = smoothing_lower_count
        assert self.smoothing_upper_count > self.smoothing_lower_count, 'smoothing upper count must be greater than lower count'
        
        self.inner_idx = jnp.array([[i, j] for i in range(1, self.height-1) for j in range(1, self.width-1)])
        self.outer_idx = jnp.array([[i, j] for i in range(self.height) for j in range(self.width) if (i==0) or (i==self.height-1) or (j==0) or (j==self.width-1)])
    
    @partial(jax.jit, static_argnums=[0])
    def sample_test_case(self, rng: chex.PRNGKey):
        
        return self.sample_barn_test_case(rng)
    
    @partial(jax.jit, static_argnums=[0])
    def sample_barn_test_case(self, rng):
        
        def _cond_fun(val):
            valid, test_case, rng = val
            # jax.debug.print('valid {v}', v=valid)
            return ~valid
        
        def _body_fun(val):
            valid, test_case, rng = val
            rng, _rng = jax.random.split(rng)
            valid, test_case = self.barn_test_case(_rng)
            return (valid, test_case, rng)
        
        init_test_case = (jnp.zeros((2,)), jnp.zeros((2,)), jnp.zeros((self.height, self.width), dtype=jnp.int32))
        
        test_case = jax.lax.while_loop(
            _cond_fun, _body_fun, (False, init_test_case, rng)
        )
        valid, test_case, rng = test_case
        
        start, goal, smoothed_map = test_case
        
        theta = (jnp.pi/2) * jax.random.choice(rng, jnp.arange(4), (2,))
        test_case = jnp.vstack([start, goal])
        test_case = jnp.concatenate([test_case, theta[:,None]], axis=1)
        
        return smoothed_map, test_case[None]
        
    def barn_test_case(self, rng):
        # from matplotlib import pyplot as plt
        rng, _rng = jax.random.split(rng)
        base_map = self.sample_map(_rng)
        
        def _smooth_fn(map_data, unused):
            inner_map = jax.vmap(self._smooth, in_axes=(0, None))(self.inner_idx, map_data)
            inner_map = inner_map.reshape((self.height-2, self.width-2))
            map_data = map_data.at[1:-1, 1:-1].set(inner_map)
            return map_data, None
                
        smoothed_map, _ = jax.lax.scan(
            _smooth_fn, base_map, None, length=self.smoothing_iters
        )        
        
        def _inflate_obs(idx, map):
            value = map.at[idx[0], idx[1]].get()
            around = map.at[INFLATE_IDX_OFFSETS_5[:,0]+idx[0], INFLATE_IDX_OFFSETS_5[:,1]+idx[1]].set(value)
            return around
        
        inner_inflated_map = jax.vmap(_inflate_obs, in_axes=(0, None))(self.inner_idx, smoothed_map).any(axis=0)
        outer_inflated_map = jax.vmap(_inflate_obs, in_axes=(0, None))(self.outer_idx, smoothed_map).any(axis=0)
        inflated_map = inner_inflated_map | outer_inflated_map
        # print('inflated map\n', inflated_map)
        # print('valid cells', (1-inflated_map).sum())
        
        rng, _rng = jax.random.split(rng)
        start = jax.random.choice(rng, jnp.arange(self.height*self.width), p=~inflated_map.flatten())
        start = jnp.array([start % self.width, start // self.width])  # [x, y]
        # print('start:', start)
        # fig, ax = plt.subplots()
        # inflated_to_plot = inflated_map * 0.1 + smoothed_map * 0.9
        # ax.imshow(inflated_to_plot, cmap='binary')
        # plt.savefig('barn-inflated-map.png')
        
        
        # print('start_idx:', start)
        # with jax.disable_jit(False):
        connected_region = _graph_utils.component_mask_with_pos(inflated_map, start)


        min_dist = 19 
        # empty = jnp.zeros((self.height, self.width), dtype=jnp.int32)
        x_lim = jnp.clip(jnp.array([start[0]-min_dist, start[0]+min_dist+1]), 0, self.width)
        y_lim = jnp.clip(jnp.array([start[1]-min_dist, start[1]+min_dist+1]), 0, self.height)
        print('x_lim:', x_lim, 'y_lim:', y_lim)

        too_close_mask = jnp.ones((self.height, self.width))
        xrange = jnp.arange(self.width)
        too_close_mask_x = jnp.where((xrange >= x_lim[0]) & (xrange < x_lim[1]), 1, 0)
        yrange = jnp.arange(self.width)
        too_close_mask_y = jnp.where((yrange >= y_lim[0]) & (yrange < y_lim[1]), 1, 0)
        
        too_close_mask = jnp.meshgrid(too_close_mask_x, too_close_mask_y)
        valid_mask = 1-jnp.dstack(too_close_mask).all(axis=-1)
        # print('too_close_mask:', valid_mask, valid_mask.shape)
        
        masked_connected_region = connected_region * valid_mask
        
        # fig, ax = plt.subplots()
        # ax.imshow(masked_connected_region, cmap='binary')
        # plt.savefig('barn-masked-map.png')
        
        goal = jax.random.choice(rng, jnp.arange(self.height*self.width), p=masked_connected_region.flatten())
        goal = jnp.array([goal % self.width, goal // self.width])  # [x, y]
        
        
        # fig, ax = plt.subplots()
        # ax.imshow(inflated_to_plot, cmap='binary')
        # ax.plot([start[0], goal[0]], [start[1], goal[1]], c='blue', linestyle='--')
        # ax.scatter(start[0], start[1], c='red', marker='x')
        # # ax.scatter(goal % cols, goal // cols, c='green', marker='x')
        # ax.scatter(goal[0], goal[1], c='green', marker='x')
        # plt.savefig('barn-final-map.png')
        
        valid = ((1-inflated_map).sum() > 0) & (masked_connected_region.sum() > 0)
        # print('test case', (start, goal))
        return valid, (start * self.cell_size, goal * self.cell_size, smoothed_map)

        
    def _smooth(self, idx, map_data):
        
        idx_offsets = SMOOTHING_IDX_OFFSETS + idx
        valid = (idx_offsets[:, 0] > 0) & (idx_offsets[:, 0] < self.height-1) & (idx_offsets[:, 1] > 0) & (idx_offsets[:, 1] < self.width-1)
        n_full = map_data.at[idx_offsets[:,0], idx_offsets[:,1]].get() * valid
        n_full = jnp.sum(n_full)
    
        fill = ((n_full>self.smoothing_upper_count) | map_data.at[idx[0], idx[1]].get()) & (n_full>self.smoothing_lower_count)
        return jax.lax.select(fill, 1, 0)
    
class GridMapPolygonAgentsSingleMap(GridMapPolygonAgents):
    
    def __init__(self,
                 num_agents: int,
                 rad,
                 map_data: List,
                 agent_coords=default_coords,
                 middle_line=middle_line,
                 **map_kwargs):
        
        self._map_data = jnp.array(
            [[int(x) for x in row.split()] for row in map_data], 
            dtype=jnp.int32
        )
        height, width = self._map_data.shape
        map_size = (width, height)
        super().__init__(num_agents=num_agents,
                        rad=rad,
                        map_size=map_size,
                        agent_coords=agent_coords,
                        middle_line=middle_line,
                        **map_kwargs)

    @partial(jax.jit, static_argnums=[0])
    def sample_map(self, key):
        return self._map_data

class GridMapFromBuffer(GridMapCircleAgents):
    
    def __init__(self,
                 num_agents,
                 rad,
                 map_size,
                 map_grids=None,
                 dir_path=None,
                 file_prefix="map_buffer_"):
        """ 
        saved map buffers expected in format: (map_data, starts, theta, goals)
        """
        print('** Super old code beware **')
        super().__init__(num_agents, rad, map_size, fill=0.1)
        if map_grids is None and dir_path is None:
            raise ValueError("Must specify either map_grids or dir_path")
        if map_grids is not None and dir_path is not None:
            raise ValueError("Cannot specify both map_grids and dir_path")
        
        if dir_path is not None:
            # list files in dir_path 
            files = [filename for filename in os.listdir(dir_path) if filename.startswith(file_prefix)]
            print('files', files)
            test_cases = (
                jnp.empty((0, self.height, self.width), dtype=jnp.int32), 
                jnp.empty((0, 2), dtype=jnp.float32),
                jnp.empty((0, 1), dtype=jnp.float32),
                jnp.empty((0, 2), dtype=jnp.float32),
            )
            for filename in files:
                # load pkls
                filepath = os.path.join(dir_path, filename)
                with open(filepath, "rb") as f:
                    tc = pickle.load(f)
                    print('tc c', tc)
                    test_cases = jax.tree.map(lambda x, y: jnp.concatenate((x, y), axis=0), test_cases, tc)
            self.test_cases = test_cases
            self.num_test_cases = test_cases[0].shape[0]
            print('test cases', test_cases)
     
        if map_grids is not None:
            raise NotImplementedError("map_grids not implemented yet")
        
    @partial(jax.jit, static_argnums=[0])
    def sample_scenario(self, key):
        print('-- sampling scenarios -- ')
        idx = jax.random.randint(key, (1,), minval=0, maxval=self.num_test_cases)[0]
        tc = jax.tree.map(lambda x: x[idx], self.test_cases)
        print('tc ', tc)
        map_data = tc[0]
        print('map data', map_data.shape)
        
        theta = jnp.array([tc[2], 0])
        print('theta', theta)
        
        case = jnp.array([tc[1], tc[3]])
        print('case', case)
        test_case = jnp.concatenate([case, theta], axis=-1)
        
        return map_data, test_case
    
    '''@partial(jax.jit, static_argnums=[0])
    def sample_map(self, key):
        """ Sample map grid from pre-specified map grids list """
        if self.map_grids.shape[0]>1:
            map_idx = jax.random.randint(key, (1,), minval=0, maxval=len(self.map_grids))[0]
            map_grid = self.map_grids[map_idx]
        else:
            map_grid = self.map_grids[0]
        return map_grid'''
    
def rrt_reward(new_pos, pos, goal):
    goal_reached = jnp.linalg.norm(new_pos - goal) <= 0.3
    #if goal_reached: print('goal reached')
    weight_g = 0.2
    goal_rew = 1
    rga = weight_g * (jnp.linalg.norm(pos - goal) - jnp.linalg.norm(new_pos - goal))
    rg = jnp.where(goal_reached, goal_rew, rga)
    return rg
    
if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    
    key = jax.random.PRNGKey(3) # 3, 7, 9
    
    file_path = "/home/alex/repos/jax-multirobsim/failure_maps/cosmic-waterfall-17"
    map_gen = GridMapFromBuffer(1, 0.3, (7, 7), dir_path=file_path)
    
    
    s = map_gen.sample_scenario(key)    
    raise
    key, key_rrt = jax.random.split(key)
    
    map_gen = GridMapCircleAgents(1, 0.3, (10, 10), 0.5)
    map_data, case = map_gen.sample_scenario(key)
    
    start = case[:, 0, :2].flatten()
    goal = case[:, 1, :2].flatten()
    print('case', case, 'start', start, 'goal', goal)
    
    '''gr, parent = map_gen.a_star(map_data, start, goal)
    print('parent', parent)
    
    fig, ax = plt.subplots()
    
    ax.imshow(map_data, extent=(0, map_data.shape[1], 0, map_data.shape[0]), origin="lower", cmap='binary', alpha=0.8)
    
    zero_grid = np.zeros((10, 10))
    x, y = jnp.meshgrid(jnp.arange(map_data.shape[0]), jnp.arange(map_data.shape[1]))#.reshape(-1, 2)
    coords = jnp.dstack((y.flatten(), x.flatten())).squeeze()
    for i in range(parent.shape[0]):
        if parent[i] == -1: continue
        node = coords[parent[i]]
        print('node', node)
        zero_grid[node[0], node[1]] = 1
    
    ax.imshow(zero_grid, extent=(0, 10, 0, 10), origin="lower", cmap='binary', alpha=0.2)
    
    map_gen.plot_pose(ax, case)
    plt.show()
    
    raise'''
    #print('map_data', map_data)
    tree, goalr = map_gen.rrt_star(key_rrt, map_data, start, goal)
    print('tree', tree, 'goalr', goalr)
    print('case', case)
    
    fig, ax = plt.subplots()
    
    map_gen.plot(ax, map_data)
    
    #ax.scatter(case[:, 0, 0], case[:, 0, 1], c='r')
    for n in range(tree.shape[0]):
        if tree[n, 0] == 0.0: break
        ax.scatter(tree[n, 0], tree[n, 1], c='gray')
        pi = tree[n, 2]
        if pi == -1: continue
        pi = int(pi)
        ax.plot([tree[n, 0], tree[pi, 0]], [tree[n, 1], tree[pi, 1]], c='gray', marker='+', alpha=0.75)
        
    

    if goalr:
        goal_idx = jnp.argwhere(tree[:,-1]==1)
        print('goal_idx', goal_idx)
        
        for g_idx in goal_idx:
            c_idx = g_idx[0]
            path_length = 0.0
            rew = 0.0
            while c_idx != 0:
                print('cidx', c_idx, 'tree row', tree[c_idx])
                c_pos = tree[c_idx, :2]
                p_idx = int(tree[c_idx, 2])
                p_pos = tree[p_idx, :2]
                path_length += jnp.linalg.norm(c_pos - p_pos)
                rew += rrt_reward(c_pos, p_pos, goal)
                ax.plot([c_pos[0], p_pos[0]], [c_pos[1], p_pos[1]], c='r', alpha=0.25)
                print('p_pos', p_pos, 'c_pos', c_pos, 'rew', rew)
                c_idx = p_idx
            print('path_length', path_length, 'rew', rew)
    
    map_gen.plot_pose(ax, case)
    plt.show()
    '''raise
    p_idx = parent[goal_idx]
    print('p_idx', p_idx)
    while p_idx != -1:
        print('corrds', coords[p_idx])
        ax.plot( [coords[p_idx, 1]+0.5, coords[goal_idx, 1]+0.5], [coords[p_idx, 0]+0.5, coords[goal_idx, 0]+0.5], c='r')
        goal_idx = p_idx
        p_idx = parent[goal_idx]

    
    plt.show()
    
    raise
    pos = jnp.array([1.5, 1.5])
    x = map_gen.check_agent_collision(pos, map_data)
    print(x)
    print(map_data)'''
    
    


""" 
NOT USED IN A LONG TIME

collision code drawn from: https://www.jeffreythompson.org/collision-detection/poly-circle.php
"""

import jax
import jax.numpy as jnp
from functools import partial

from .map import Map
from jaxmarl.environments.jaxnav.jaxnav_utils import rot_mat

class PolygonMap(Map):
    
    def __init__(self, 
                 num_agents,
                 rad,
                 map_size,
                 num_sides=4,
                 max_num_shapes=40,
                 min_edge_length=0.2,
                 max_edge_length=4.0): 
        super().__init__(num_agents, rad, map_size)
        print('Super old code beware')
        self.num_sides = num_sides
        self.max_num_shapes = max_num_shapes
        self.min_edge_length = min_edge_length
        self.max_edge_length = max_edge_length
        
        if self.num_sides != 4:
            raise NotImplementedError('Only 4-sided polygons are currently supported')
        self.shape_fn = self._sample_parrallelogram
       
    ### === MAP GENERATION === ### 
    @partial(jax.jit, static_argnums=[0])
    def sample_map(self, key):
        """ Sample polygon map, returns coordinates of vertices """
        key_num, key_coords = jax.random.split(key)
    
        # rectangle for map bounds
        bounds = jnp.array([
            [[0.0, 0.15], [self.map_size[0], 0.15], [self.map_size[0], 0.0], [0.0, 0.0]],
            [[0.15, 0.0], [0.15, self.map_size[1]], [0.0, self.map_size[1]], [0.0, 0.0]],
            [[0.0, self.map_size[1]], [self.map_size[0], self.map_size[1]], [self.map_size[0], self.map_size[1]-0.15], [0.0, self.map_size[1]-0.15]],
            [[self.map_size[0]-0.15, 0.0], [self.map_size[0]-0.15, self.map_size[1]], [self.map_size[0],self.map_size[1]], [self.map_size[0], 0.0]],
        ])
        
        
        # out of bounds constant
        oobounds = jnp.full((self.num_sides, 2), 1.0, dtype=jnp.float32) + jnp.array([self.map_size[0], self.map_size[1]])
        
        num_shapes = jax.random.randint(key_num, (1,), minval=0, maxval=self.max_num_shapes)
        p_idx = jnp.arange(0, self.max_num_shapes)
        p_mask = jnp.where(p_idx<=num_shapes, 0, 1)
            
        key_coords = jax.random.split(key_coords, self.max_num_shapes)
        coords = self.shape_fn(key_coords)
        
        mask = p_mask[:, None, None] * jnp.repeat(oobounds[None], self.max_num_shapes, axis=0)
        coords = coords + mask
        return jnp.append(coords, bounds, axis=0)
    
    ### === COLLISION DETECTION === ###
    @partial(jax.jit, static_argnums=[0])
    def check_agent_map_collision(self, pos, coords):
        """ Check collision between a line and a circle subject to a map boundary """
        #print('agent check coords shape', coords.shape)

        @partial(jax.vmap, in_axes=(0, 0, None))
        def _check_ac(p1, p2, pos):
        
            def _in_map(p1, p2, pos, r):
                return (p1[0] >= 0) & (p1[0] <= self.map_size[0]) & (p1[1] >= 0) & (p1[1] <= self.map_size[1]) & \
                    (p2[0] >= 0) & (p2[0] <= self.map_size[0]) & (p2[1] >= 0) & (p2[1] <= self.map_size[1])
                
            return jax.lax.cond(_in_map(p1, p2, pos, self.rad), lambda _: line_circle_collision(p1, p2, pos, self.rad), lambda _: False, None)
    
        edges = self.gen_edges(coords)
        return jnp.any(_check_ac(edges[:, 0], edges[:, 1], pos))
    
    @partial(jax.jit, static_argnums=[0])
    def check_map_collision(self, pos, coords, radius):
        """ For a circle agent """
        l = self.check_agent_map_collision(pos, coords)
        i = jnp.any(self.check_point_map_collision(pos, coords))
        return l | i
    
    # NOTE need to add ray tracing for lidar - constrains size of usable polygon
    
    @partial(jax.jit, static_argnums=[0])  
    def check_point_map_collision(self, pos, coords):
        """ Check if a point is inside a polygon within a map, 
        NOTE: all map coords should be passed"""
        
        @partial(jax.vmap, in_axes=[None, 0, 0])
        def _vc(pos, vc, vn):
            c = (((vc[1] > pos[1]) & (vn[1] < pos[1])) |\
                ((vc[1] < pos[1]) & (vn[1] > pos[1]))) & \
                (pos[0] < (vn[0] - vc[0]) * (pos[1] - vc[1]) / (vn[1] - vc[1]) + vc[0])

            return c 
        
        coords = jnp.concatenate((coords, coords[:, 0][:, None]), axis=1)
        return jnp.sum(_vc(pos, coords[:, :-1].reshape(-1, 2), coords[:, 1:].reshape(-1, 2))) % 2 != 0
         
    ### === UTILS === ### 
    def gen_edges(self, coords):
        """ Generate edges from coordinates and flatten array"""
        coordsp = jnp.append(coords, coords[:, 0].reshape(-1, 1, 2), axis=1)
        return jnp.column_stack((coordsp[:, :-1], coordsp[:, 1:])).reshape((-1, 2, 2))
    
    
    @partial(jax.vmap, in_axes=[None, 0])
    def _sample_parrallelogram(self, key):
        """ Sample a parralelogram from a uniform distribution """
        
        # Sample parralelogram parameters
        key_c, key_o, key_s, key_a = jax.random.split(key, 4)
        centre = jax.random.uniform(key_c, (1, 2), minval=0.0, maxval=self.map_size[1])
        orientation = jax.random.uniform(key_o, (1,), minval=-jnp.pi, maxval=jnp.pi)
        a, b = jax.random.uniform(key_s, (2,), minval=self.min_edge_length, maxval=self.max_edge_length)
        alpha = jax.random.uniform(key_a, (1,), minval=0.1, maxval=jnp.pi)
        
        # Calculate diagonals & angles
        q = jnp.sqrt(a**2 + b**2 + 2*a*b*jnp.cos(alpha))
        p = jnp.sqrt(a**2 + b**2 - 2*a*b*jnp.cos(alpha))
        beta = jnp.arcsin(jnp.sin(alpha)*a/p)
        gamma = jnp.arcsin(jnp.sin(beta)/(q/2)*(p/2))
        
        # Calculate coordinates
        d = - p/2 * jnp.array([jnp.cos(beta), jnp.sin(-beta)]).reshape((1,2))
        b = + p/2 * jnp.array([jnp.cos(-beta), jnp.sin(-beta)]).reshape((1,2))
        c = + q/2 * jnp.array([jnp.cos(gamma), jnp.sin(gamma)]).reshape((1,2))
        a = - q/2 * jnp.array([jnp.cos(gamma), jnp.sin(gamma)]).reshape((1,2))

        return jnp.dot(jnp.array([a, b, c, d]), rot_mat(orientation)).squeeze() + centre

    ### === VISUALISATION === ###
    def plot_map(self, ax, coord):
        coord = jnp.append(coord, coord[:, 0].reshape(-1, 1, 2), axis=1)
        for c in coord:
            xs, ys = zip(*c)    
            ax.plot(xs, ys, color='black')


### === COLLISION DETECTION UTILS === ### NOTE likely a better file for these
@partial(jax.vmap, in_axes=(0, 0, None, None, None))
def line_circle_map_collision(p1, p2, c, r, map_size):
    """ Check collision between a line and a circle subject to a map boundary """
    
    def _in_map(p1, p2, c, r):
        return (p1[0] >= 0) & (p1[0] < map_size[0]) & (p1[1] >= 0) & (p1[1] < map_size[1]) & \
               (p2[0] >= 0) & (p2[0] < map_size[0]) & (p2[1] >= 0) & (p2[1] < map_size[1])
               
    return jax.lax.cond(_in_map(p1, p2, c, r), lambda _: line_circle_collision(p1, p2, c, r), lambda _: False, None)
    

def line_circle_collision(p1, p2, c, r):
    """ Check collision between a line and a circle """
    
    cc = point_circle_collision(p1, c, r) | point_circle_collision(p2, c, r)
    
    def _on_segement(p1, p2, c, r):
        d1 = jnp.linalg.norm(p1 - c)
        d2 = jnp.linalg.norm(p2 - c)
        dl = jnp.linalg.norm(p1 - p2)
        
        return (d1+d2 >= dl-r) & (d1+d2 <= dl+r)
    
    def _line_circle(p2, p1, c, r):
    
        dx, dy = p2 - p1
        len = jnp.sqrt(dx**2 + dy**2)
        dot = ((c[0] - p1[0]) * dx + (c[1] - p1[1]) * dy) / len**2
        closest = jnp.array([p1[0] + dot * dx, p1[1] + dot * dy])
        
        # check if closest on line segment
        on_seg = _on_segement(p1, p2, c, r)
        
        dist = closest - c
        return (jnp.sqrt(dist[0]**2 + dist[1]**2) <= r) & (on_seg)
    
    return jax.lax.cond(cc, lambda _: True, lambda _: _line_circle(p1, p2, c, r), None)

def point_circle_collision(p, c, r):
    """ Check collision between a point and a circle """
    dist = p - c
    return jnp.sqrt(dist[0]**2 + dist[1]**2) <= r    


if __name__=="__main__":
    key = jax.random.PRNGKey(10)
    
    map_size = (10, 10)
    rad = 0.3
    
    map_gen = PolygonMap(2, rad, map_size)
    map_coords = map_gen.sample_map(key)
    
    print('map coords', map_coords[0])
    
    pos = jnp.array([5.3, 6.5])
    #c = map_gen.check_point_collision(pos, map_coords[0])
    c = map_gen.check_agent_map_collision(pos, map_coords)
    print('c', c)
    c = map_gen.check_map_collision(pos, map_coords, None)
    print('c', c)
    
    import matplotlib.pyplot as plt
    # from jax_multirobsim.env.sample_cases.create_sample_cases import jax_sample_case
    
    # case = jax_sample_case(key, 2, 0.3, map_size, map_coords, map_fn=map_gen.check_map_collision)
    
    # fig, ax = plt.subplots()
    
    # map_gen.plot_map(ax, map_coords)
    # #plot_sample_case()
    # ax.scatter(pos[0], pos[1], color='red')
    
    # plt.show()
    
    
    


""" Built off Gymnax spaces.py, this module contains jittable classes for action and observation spaces. """
from typing import Tuple, Union, Sequence
from collections import OrderedDict
import chex
import jax
import jax.numpy as jnp

class Space(object):
    """
    Minimal jittable class for abstract jaxmarl space.
    """

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        raise NotImplementedError

    def contains(self, x: jnp.int_) -> bool:
        raise NotImplementedError

class Discrete(Space):
	"""
	Minimal jittable class for discrete gymnax spaces.
	TODO: For now this is a 1d space. Make composable for multi-discrete.
	"""

	def __init__(self, num_categories: int, dtype=jnp.int32):
		assert num_categories >= 0
		self.n = num_categories
		self.shape = ()
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from set of categorical choices."""
		return jax.random.randint(
			rng, shape=self.shape, minval=0, maxval=self.n
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(x >= 0, x < self.n)
		return range_cond


class MultiDiscrete(Space):
    """
    Minimal jittable class for multi-discrete gymnax spaces.
    """

    def __init__(self, num_categories: Sequence[int]):
        """Num categories is the number of cat actions for each dim, [2,2,2]=2 actions x 3 dim"""
        self.num_categories = jnp.array(num_categories)
        self.shape = (len(num_categories),)
        self.dtype = jnp.int_

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        """Sample random action uniformly from set of categorical choices."""
        return jax.random.randint(
            rng, 
            shape=self.shape, 
            minval=0, 
            maxval=self.num_categories,
            dtype=self.dtype
        )

    def contains(self, x: jnp.int_) -> bool:
        """Check whether specific object is within space."""
        range_cond = jnp.logical_and(x >= 0, x < self.num_categories)
        return jnp.all(range_cond)


class Box(Space):
	"""
	Minimal jittable class for array-shaped gymnax spaces.
	TODO: Add unboundedness - sampling from other distributions, etc.
	"""
	def __init__(
		self,
		low: float,
		high: float,
		shape: Tuple[int],
		dtype: jnp.dtype = jnp.float32,
	):
		self.low = low
		self.high = high
		self.shape = shape
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from 1D continuous range."""
		return jax.random.uniform(
			rng, shape=self.shape, minval=self.low, maxval=self.high
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(
			jnp.all(x >= self.low), jnp.all(x <= self.high)
		)
		return range_cond


class Dict(Space):
	"""Minimal jittable class for dictionary of simpler jittable spaces."""
	def __init__(self, spaces: dict):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> dict:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return OrderedDict(
			[
				(k, self.spaces[k].sample(key_split[i]))
				for i, k in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, dict)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for k, space in self.spaces.items():
			out_of_space += 1 - space.contains(getattr(x, k))
		return out_of_space == 0


class Tuple(Space):
	"""Minimal jittable class for tuple (product) of jittable spaces."""
	def __init__(self, spaces: Union[tuple, list]):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> Tuple[chex.Array]:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return tuple(
			[
				space.sample(key_split[i])
				for i, space in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, tuple)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for space in self.spaces:
			out_of_space += 1 - space.contains(x)
		return out_of_space == 0
